@@ -22,14 +22,13 @@ extern char *findpattern;
 extern int sorttype;
 extern int writefs;
 extern int minfilename;
+extern char *curfilename;
 
 static VOID pathbar();
 static VOID stackbar();
 static char *putowner();
 static char *putgroup();
-static VOID infobar();
 static VOID calclocate();
-static VOID putname();
 static int browsedir();
 
 int columns;
@@ -42,7 +41,6 @@ int chgorder;
 int stackdepth = 0;
 namelist filestack[MAXSTACK];
 char fullpath[MAXPATHLEN + 1];
-char *lastpath = NULL;
 char *origpath;
 char *macrolist[MAXMACROTABLE];
 int maxmacro = 0;
@@ -51,8 +49,9 @@ char *helpindex[10] = {
 	"Sort", "Find", "Tree", "Editor", "Unpack"
 };
 
-static namelist *filelist;
-static int maxfile;
+namelist *filelist;
+int maxfile;
+int maxent;
 
 
 static VOID pathbar()
@@ -241,7 +240,7 @@ gid_t gid;
 	return(buf);
 }
 
-static VOID infobar(list, no)
+VOID infobar(list, no)
 namelist *list;
 int no;
 {
@@ -297,7 +296,7 @@ int no;
 		}
 		if (len < n_lastcolumn) {
 			width = n_lastcolumn - len;
-			readlink(list[no].name, buf + len, width);
+			Xreadlink(list[no].name, buf + len, width);
 		}
 	}
 
@@ -305,6 +304,15 @@ int no;
 	tflush();
 }
 
+VOID waitmes()
+{
+	helpbar();
+	locate(0, LMESLINE);
+	putterm(l_clear);
+	kanjiputs(WAIT_K);
+	tflush();
+}
+
 static VOID calclocate(i)
 int i;
 {
@@ -331,10 +339,9 @@ int calcwidth()
 	return(width);
 }
 
-static VOID putname(list, no, standout)
+VOID putname(list, no, standout)
 namelist *list;
-int no;
-int standout;
+int no, standout;
 {
 	char buf[MAXLINESTR + 1];
 	struct tm *tm;
@@ -351,31 +358,28 @@ int standout;
 
 	width = calcwidth();
 	i = (standout && fnameofs > 0) ? fnameofs : 0;
-	strncpy3(buf, list[no].name, width, i);
-	if (isdisptyp(dispmode)) {
-		for (i = 0; i < width; i++) if (buf[i] == ' ') {
-			switch (list[no].st_mode & S_IFMT) {
-				case S_IFDIR:
-					buf[i] = '/';
-					break;
-				case S_IFLNK:
-					buf[i] = '@';
-					break;
-				case S_IFSOCK:
-					buf[i] = '=';
-					break;
-				case S_IFIFO:
-					buf[i] = '|';
-					break;
-				case S_IFREG:
-					if (list[no].st_mode &
-					(S_IXUSR | S_IXGRP | S_IXOTH))
-						buf[i] = '*';
-					break;
-				default:
-					break;
-			}
-			break;
+	i = strncpy3(buf, list[no].name, width, i);
+	if (isdisptyp(dispmode) && i < width) {
+		switch (list[no].st_mode & S_IFMT) {
+			case S_IFDIR:
+				buf[i] = '/';
+				break;
+			case S_IFLNK:
+				buf[i] = '@';
+				break;
+			case S_IFSOCK:
+				buf[i] = '=';
+				break;
+			case S_IFIFO:
+				buf[i] = '|';
+				break;
+			case S_IFREG:
+				if (list[no].st_mode &
+				(S_IXUSR | S_IXGRP | S_IXOTH))
+					buf[i] = '*';
+				break;
+			default:
+				break;
 		}
 	}
 	len = width;
@@ -384,6 +388,9 @@ int standout;
 	if (columns < 5 && len + WIDTH3 <= width) {
 		if (isdir(&list[no]))
 			sprintf(buf + len, " %*.*s", WSIZE, WSIZE, "<DIR>");
+		else if (dospath("", NULL)
+		&& (list[no].st_mode & S_IFMT) == S_IFIFO)
+			sprintf(buf + len, " %*.*s", WSIZE, WSIZE, "<VOL>");
 		else if (isdev(&list[no])) sprintf(buf + len, " %*u,%*u",
 			WSIZE / 2, ((unsigned)(list[no].st_size) >> 8) & 0xff,
 			WSIZE - (WSIZE / 2) - 1,
@@ -493,20 +500,23 @@ u_char fstat;
 	infobar(list, filepos);
 }
 
-VOID rewritefile()
+VOID rewritefile(all)
+int all;
 {
 	pathbar();
-	helpbar();
+	if (all) {
+		helpbar();
+		infobar(filelist, filepos);
+	}
 	statusbar(maxfile);
 	stackbar();
-	movepos(filelist, maxfile, filepos, LISTUP);
+	listupfile(filelist, maxfile, filelist[filepos].name);
 	locate(0, 0);
 	tflush();
 }
 
-static int browsedir(file, def, maxentp)
+static int browsedir(file, def)
 char *file, *def;
-int *maxentp;
 {
 	DIR *dirp;
 	struct dirent *dp;
@@ -519,13 +529,9 @@ int *maxentp;
 	chgorder = 0;
 	if (sorttype < 100) sorton = sorttype;
 
-	if (!(dirp = opendir("."))) error(".");
+	if (!(dirp = Xopendir("."))) error(".");
 	fnameofs = 0;
-
-	locate(0, LMESLINE);
-	putterm(l_clear);
-	kanjiputs(WAIT_K);
-	tflush();
+	waitmes();
 
 	if (!findpattern) re = NULL;
 	else {
@@ -543,30 +549,32 @@ int *maxentp;
 		if (i < stackdepth) continue;
 
 		filelist = (namelist *)addlist(filelist, maxfile,
-			maxentp, sizeof(namelist));
+			&maxent, sizeof(namelist));
+		if (getstatus(filelist, maxfile, file) < 0) continue;
 		filelist[maxfile].name = strdup2(file);
 		filelist[maxfile].ent = maxfile;
-		getstatus(filelist, maxfile, file);
 		maxfile++;
 	}
-	closedir(dirp);
+	Xclosedir(dirp);
 	if (re) regexp_free(re);
 
-	if (maxfile <= 0) {
+	if (!maxfile) {
 		filelist = (namelist *)addlist(filelist, 0,
-			maxentp, sizeof(namelist));
-		filelist[0].name = strdup2(NOFIL_K);
+			&maxent, sizeof(namelist));
+		filelist[0].name = NOFIL_K;
 		filelist[0].st_nlink = -1;
 	}
 
 	if (sorton) qsort(filelist, maxfile, sizeof(namelist), cmplist);
 
-	if (stable_standout) putterms(t_clear);
+	if (stable_standout) {
+		putterms(t_clear);
+		helpbar();
+	}
 	title();
 	pathbar();
 	statusbar(maxfile);
 	stackbar();
-	helpbar();
 
 	old = filepos = listupfile(filelist, maxfile, def);
 	fstat = 0;
@@ -577,7 +585,9 @@ int *maxentp;
 		if (no) movepos(filelist, maxfile, old, fstat);
 		locate(0, 0);
 		tflush();
-		ch = getkey(SIGALRM);
+		getkey2(-1);
+		ch = getkey2(SIGALRM);
+		getkey2(-1);
 
 		old = filepos;
 		for (i = 0; i < MAXBINDTABLE && bindlist[i].key >= 0; i++)
@@ -605,7 +615,7 @@ int *maxentp;
 				stackbar();
 			}
 #if (WRITEFS < 1)
-			else if (chgorder && writefs < 1
+			else if (chgorder && writefs < 1 && no != WRITE_DIR
 			&& (i = writablefs(".")) > 0 && underhome() > 0) {
 				chgorder = 0;
 				if (yesno(WRTOK_K))
@@ -614,6 +624,7 @@ int *maxentp;
 #endif
 		}
 
+		curfilename = filelist[filepos].name;
 		if (maxfile <= 0 && !(fstat & NO_FILE)) no = 0;
 		else if (no <= NO_OPERATION)
 			no = (*funclist[no].func)(filelist, &maxfile);
@@ -655,7 +666,7 @@ char *cur;
 {
 	char file[MAXNAMLEN + 1], prev[MAXNAMLEN + 1];
 	char *cp, *def;
-	int ischgdir, maxent;
+	int i, ischgdir;
 
 	origpath = getwd2();
 	strcpy(fullpath, origpath);
@@ -663,26 +674,30 @@ char *cur;
 	findpattern = def = NULL;
 	if (cur) {
 		cp = evalpath(strdup2(cur));
-		if (chdir(cp) >= 0) {
-			if (*cp == '/') *fullpath = '\0';
-			else strcat(fullpath, "/");
+		if (_dospath(cp) || *cp == '/') *fullpath = '\0';
+		else strcat(fullpath, "/");
+		if (_chdir2(cp) >= 0) {
 			strcat(fullpath, cp);
 			free(cp);
 			cp = fullpath + strlen(fullpath) - 1;
-			if (cp > fullpath && *cp == '/') *cp = '\0';
 		}
 		else {
 			def = strrchr(cp, '/');
+			if (!def && _dospath(cp)) def = &cp[2];
 			if (def) {
+				i = *def;
 				*def = '\0';
-				if (def++ == cp) {
-					if (chdir("/") < 0) error("/");
+				if (def == cp) {
+					if (_chdir2("/") < 0) error("/");
 					else strcpy(fullpath, "/");
 				}
+				else if (_chdir2(cp) >= 0) strcat(fullpath, cp);
 				else {
-					if (chdir(cp) < 0) def = NULL;
-					else strcpy(fullpath, cp);
+					warning(-1, cp);
+					def = NULL;
+					strcpy(fullpath, origpath);
 				}
+				if (def && (*def = i) == '/') def++;
 			}
 			else def = cp;
 
@@ -692,6 +707,7 @@ char *cur;
 			}
 			free(cp);
 		}
+		realpath2(fullpath, fullpath);
 	}
 
 	filelist = NULL;
@@ -721,7 +737,7 @@ char *cur;
 			}
 		}
 
-		ischgdir = browsedir(file, def, &maxent);
+		ischgdir = browsedir(file, def);
 		if (ischgdir < 0) break;
 		if (ischgdir) def = NULL;
 		else {
@@ -732,6 +748,6 @@ char *cur;
 	}
 
 	free(filelist);
-	if (chdir(origpath) < 0) error(origpath);
+	if (_chdir2(origpath) < 0) error(origpath);
 	free(origpath);
 }
