@@ -24,8 +24,8 @@ extern int chgorder;
 extern int stackdepth;
 extern int copypolicy;
 extern namelist filestack[];
-extern char fullpath[];
 extern char *archivefile;
+extern char *archivedir;
 extern char *destpath;
 extern int writefs;
 extern int savehist;
@@ -51,7 +51,9 @@ static int fname_right();
 static int fname_left();
 static int mark_file();
 static int mark_file2();
+static int mark_file3();
 static int mark_all();
+static int mark_reverse();
 static int mark_find();
 static int push_file();
 static int pop_file();
@@ -139,6 +141,7 @@ bindtable bindlist[MAXBINDTABLE] = {
 	{')',		FNAME_LEFT,	255},
 	{' ',		MARK_FILE2,	255},
 	{'+',		MARK_ALL,	255},
+	{'-',		MARK_REVERSE,	255},
 	{'*',		MARK_FIND,	255},
 	{']',		PUSH_FILE,	255},
 	{'[',		POP_FILE,	255},
@@ -175,17 +178,12 @@ bindtable bindlist[MAXBINDTABLE] = {
 	{'S',		SYMLINK_MODE,	255},
 	{'T',		FILETYPE_MODE,	255},
 	{'U',		UNPACK_TREE,	255},
-	{CTRL_A,	CUR_TOP,	255},
-	{CTRL_B,	CUR_LEFT,	255},
-	{CTRL_C,	ROLL_UP,	255},
-	{CTRL_E,	CUR_BOTTOM,	255},
-	{CTRL_F,	CUR_RIGHT,	255},
-	{CTRL_L,	REREAD_DIR,	255},
-	{CTRL_N,	CUR_DOWN,	255},
-	{CTRL_P,	CUR_UP,		255},
-	{CTRL_R,	ROLL_DOWN,	255},
-	{CTRL_V,	ROLL_UP,	255},
-	{CTRL_Y,	ROLL_DOWN,	255},
+	{K_HOME,	MARK_ALL,	255},
+	{K_END,		MARK_REVERSE,	255},
+	{K_BEG,		CUR_TOP,	255},
+	{K_EOL,		CUR_BOTTOM,	255},
+	{CTRL('@'),	MARK_FILE3,	255},
+	{CTRL('L'),	REREAD_DIR,	255},
 	{-1,		NO_OPERATION,	255}
 };
 
@@ -295,8 +293,11 @@ static int out_dir(list, maxp)
 namelist *list;
 int *maxp;
 {
-	free(list[filepos].name);
-	list[filepos].name = strdup2("..");
+	if (archivefile) filepos = -1;
+	else {
+		free(list[filepos].name);
+		list[filepos].name = strdup2("..");
+	}
 	return(5);
 }
 
@@ -361,7 +362,7 @@ int *maxp;
 		|| (list[filepos].st_mode & S_IFMT) == S_IFLNK
 		|| (list[filepos].st_mode & S_IFMT) == S_IFSOCK
 		|| (list[filepos].st_mode & S_IFMT) == S_IFIFO
-		|| access(list[filepos].name, X_OK) >= 0)) i--;
+		|| Xaccess(list[filepos].name, X_OK) >= 0)) i--;
 	}
 	if (fnameofs + i >= strlen(list[filepos].name)) return(0);
 	fnameofs++;
@@ -392,6 +393,18 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
+static int mark_file3(list, maxp)
+namelist *list;
+int *maxp;
+{
+	mark_file(list, maxp);
+	if (filepos < *maxp - 1
+	&& filepos / FILEPERPAGE == (filepos + 1) / FILEPERPAGE) filepos++;
+	else filepos = (filepos / FILEPERPAGE) * FILEPERPAGE;
+	return(2);
+}
+
 /*ARGSUSED*/
 static int mark_all(list, maxp)
 namelist *list;
@@ -415,6 +428,21 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
+static int mark_reverse(list, maxp)
+namelist *list;
+int *maxp;
+{
+	int i;
+
+	mark = 0;
+	for (i = 0; i < *maxp; i++) if (!isdir(&list[i]))
+		if (list[i].flags ^= F_ISMRK) mark++;
+	locate(17, LSTATUS);
+	cprintf("%4d", mark);
+	return(2);
+}
+
 /*ARGSUSED*/
 static int mark_find(list, maxp)
 namelist *list;
@@ -424,7 +452,7 @@ int *maxp;
 	char *cp, *wild;
 	int i;
 
-	if (!(wild = inputstr2(FINDF_K, 0, "*", NULL))) return(1);
+	if (!(wild = inputstr(FINDF_K, 0, 0, "*", NULL))) return(1);
 	if (!*wild || strchr(wild, '/')) {
 		warning(ENOENT, wild);
 		free(wild);
@@ -495,7 +523,7 @@ namelist *list;
 int *maxp;
 {
 	dispmode ^= F_FILETYPE;
-	return(3);
+	return(2);
 }
 
 /*ARGSUSED*/
@@ -514,7 +542,7 @@ int *maxp;
 {
 	char *path;
 
-	if (!(path = inputstr2(LOGD_K, -1, NULL, NULL))
+	if (!(path = inputstr(LOGD_K, 0, -1, NULL, NULL))
 	|| !*(path = evalpath(path))) return(1);
 	if (!chdir3(path)) {
 		warning(-1, path);
@@ -594,19 +622,24 @@ namelist *list;
 int *maxp;
 {
 	char *dir;
+	int drive;
 
+	drive = 0;
 	if (isdir(&list[filepos])
-	|| (archivefile && !(dir = tmpunpack(list, *maxp)))) return(0);
+	|| (archivefile && !(dir = tmpunpack(list, *maxp)))
+	|| ((drive = dospath("", NULL)) && !(dir = tmpdosdupl(drive,
+	list[filepos].name, list[filepos].st_mode)))) return(1);
 	if (!execenv("FD_PAGER", list[filepos].name)) {
 #ifdef	PAGER
-		execmacro(PAGER, list[filepos].name, NULL, 0, 0, 0);
+		execmacro(PAGER, list[filepos].name, NULL, 0, 1, 0);
 #else
 		do {
 			dump(list[filepos].name);
 		} while(!yesno(PEND_K));
 #endif
 	}
-	if (archivefile) removetmp(dir, list[filepos].name);
+	if (drive) removetmp(dir, NULL, list[filepos].name);
+	else if (archivefile) removetmp(dir, archivedir, list[filepos].name);
 	return(2);
 }
 
@@ -615,12 +648,22 @@ static int edit_file(list, maxp)
 namelist *list;
 int *maxp;
 {
+	char *dir;
+	int drive;
+
 	if (isdir(&list[filepos])) return(warning_bell(list, maxp));
+	if ((drive = dospath("", NULL)) && !(dir = tmpdosdupl(drive,
+	list[filepos].name, list[filepos].st_mode))) return(1);
 	if (!execenv("FD_EDITOR", list[filepos].name)) {
 #ifdef	EDITOR
-		execmacro(EDITOR, list[filepos].name, NULL, 0, 0, 0);
+		execmacro(EDITOR, list[filepos].name, NULL, 0, 1, 0);
 #endif
 	}
+	if (drive) {
+		if (tmpdosrestore(drive, list[filepos].name,
+		list[filepos].st_mode) < 0) warning(-1, list[filepos].name);
+		removetmp(dir, NULL, list[filepos].name);
+	}
 	return(4);
 }
 
@@ -673,7 +716,7 @@ int *maxp;
 		free(dupl);
 		chgorder = 1;
 	}
-	return(3);
+	return(2);
 }
 
 /*ARGSUSED*/
@@ -714,7 +757,7 @@ namelist *list;
 int *maxp;
 {
 	help(archivefile != NULL);
-	return(3);
+	return(2);
 }
 
 /*ARGSUSED*/
@@ -734,7 +777,7 @@ int *maxp;
 {
 	char *path;
 
-	if (!(path = inputstr2(MAKED_K, -1, NULL, NULL))
+	if (!(path = inputstr(MAKED_K, 1, -1, NULL, NULL))
 	|| !*(path = evalpath(path))) return(1);
 	if (mkdir2(path, 0777) < 0) warning(-1, path);
 	free(path);
@@ -784,17 +827,17 @@ int *maxp;
 	if (!strcmp(list[filepos].name, ".")
 	|| !strcmp(list[filepos].name, "..")) return(warning_bell(list, maxp));
 	for (;;) {
-		if (!(file = inputstr2(NEWNM_K, 0, list[filepos].name, NULL)))
+		if (!(file = inputstr(NEWNM_K, 1, 0, list[filepos].name, NULL)))
 			return(1);
 		file = evalpath(file);
-		if (lstat(file, &status) < 0) {
+		if (Xlstat(file, &status) < 0) {
 			if (errno == ENOENT) break;
 			warning(-1, file);
 		}
 		else warning(0, WRONG_K);
 		free(file);
 	}
-	if (rename(list[filepos].name, file) < 0) {
+	if (Xrename(list[filepos].name, file) < 0) {
 		warning(-1, file);
 		free(file);
 		return(1);
@@ -850,7 +893,7 @@ int *maxp;
 	}
 	else if (!applydir(list[filepos].name, unlink2, NULL, rmdir2, NULL))
 		filepos++;
-	if (filepos >= *maxp) filepos -= 2;
+	if (filepos >= *maxp && (filepos -= 2) < 0) filepos = 0;
 	return(4);
 }
 
@@ -861,7 +904,7 @@ int *maxp;
 {
 	char *wild;
 
-	if (!(wild = inputstr2(FINDF_K, 0, "*", NULL))) return(1);
+	if (!(wild = inputstr(FINDF_K, 0, 0, "*", NULL))) return(1);
 	if (strchr(wild, '/')) {
 		warning(ENOENT, wild);
 		free(wild);
@@ -883,7 +926,7 @@ int *maxp;
 {
 	char *cp, *wild;
 
-	if (!(wild = inputstr2(FINDD_K, 0, "*", NULL))) return(1);
+	if (!(wild = inputstr(FINDD_K, 0, 0, "*", NULL))) return(1);
 	if (!*wild || strchr(wild, '/')) {
 		warning(ENOENT, wild);
 		free(wild);
@@ -918,7 +961,7 @@ int *maxp;
 {
 	char *com;
 
-	if (!(com = inputstr2("sh#", -1, NULL, &sh_history))) return(1);
+	if (!(com = inputstr("sh#", 0, -1, NULL, &sh_history))) return(1);
 	if (*com) execmacro(com, list[filepos].name, list, *maxp, 0, 1);
 	else execshell();
 	free(com);
@@ -931,26 +974,34 @@ namelist *list;
 int *maxp;
 {
 	char *com, *dir;
-	int len;
+	int len, drive;
 
-	len = (access(list[filepos].name, X_OK) >= 0) ?
+	len = (Xaccess(list[filepos].name, X_OK) >= 0) ?
 		strlen(list[filepos].name) + 1 : 0;
-	if (!(com = inputstr2("sh#", len, list[filepos].name, &sh_history)))
+	if (!(com = inputstr("sh#", 0, len, list[filepos].name, &sh_history)))
 		return(1);
 	if (!*com) {
 		execshell();
 		free(com);
 		return(4);
 	}
-	if (archivefile && !(dir = tmpunpack(list, *maxp))) {
+
+	drive = 0;
+	if ((archivefile && !(dir = tmpunpack(list, *maxp)))
+	|| ((drive = dospath("", NULL)) && !(dir = tmpdosdupl(drive,
+	list[filepos].name, list[filepos].st_mode)))) {
 		free(com);
 		return(1);
 	}
-	if (!archivefile) execmacro(com, list[filepos].name, list, *maxp, 0, 1);
-	else {
+	if (drive) {
 		execmacro(com, list[filepos].name, NULL, 0, 0, 1);
-		removetmp(dir, list[filepos].name);
+		removetmp(dir, NULL, list[filepos].name);
 	}
+	else if (archivefile) {
+		execmacro(com, list[filepos].name, NULL, 0, 0, 1);
+		removetmp(dir, archivedir, list[filepos].name);
+	}
+	else execmacro(com, list[filepos].name, list, *maxp, 0, 1);
 	free(com);
 	return(4);
 }
@@ -972,7 +1023,7 @@ int *maxp;
 	char *file;
 	int i;
 
-	if (!(file = inputstr2(PACK_K, -1, NULL, NULL))
+	if (!(file = inputstr(PACK_K, 1, -1, NULL, NULL))
 	|| !*(file = evalpath(file))) return(1);
 	i = pack(file, list, *maxp);
 	free(file);
@@ -1023,7 +1074,7 @@ int *maxp;
 	char *path;
 	int i;
 
-	if (!(path = inputstr2(FSDIR_K, -1, NULL, NULL))) return(1);
+	if (!(path = inputstr(FSDIR_K, 0, -1, NULL, NULL))) return(1);
 	if (!*(path = evalpath(path))) {
 		free(path);
 		path = strdup2(".");
@@ -1054,8 +1105,7 @@ int *maxp;
 		locate(0, LINFO);
 		putterm(l_clear);
 		kanjiputs(ATTRM_K);
-		if (selectstr(&flag, 2, 35, str, val) == ESC)
-			return(1);
+		if (selectstr(&flag, 2, 35, str, val) == ESC) return(1);
 	}
 	else {
 		if (islink(&list[filepos])) {
@@ -1067,7 +1117,7 @@ int *maxp;
 	}
 
 	while ((n = inputattr(&list[i], flag)) < 0) warning(0, ILTMS_K);
-	if (!n) return(3);
+	if (!n) return(2);
 
 	if (mark > 0) applyfile(list, *maxp, setattr, NULL);
 	else if (setattr(list[filepos].name) < 0)
@@ -1082,11 +1132,11 @@ int *maxp;
 {
 	char *path;
 
-	if (!(path = tree(0))) return(3);
+	if (!(path = tree(0, NULL))) return(2);
 	if (chdir2(path) < 0) {
 		warning(-1, path);
 		free(path);
-		return(3);
+		return(2);
 	}
 	free(path);
 	if (findpattern) free(findpattern);
@@ -1104,7 +1154,7 @@ int *maxp;
 	char *dev;
 	int i;
 
-	if (!(dev = inputstr2(BKUP_K, 5, "/dev/", NULL))
+	if (!(dev = inputstr(BKUP_K, 1, 5, "/dev/", NULL))
 	|| !*(dev = evalpath(dev))) return(1);
 	i = backup(dev, list, *maxp);
 	free(dev);
