@@ -30,8 +30,8 @@
  *	If you don't like the following tools as each uses,	*
  *	you should rewrite another suitable command name.	*
  ****************************************************************/
-#define PAGER		"more"		/* to view file */
-#define EDITOR		"vi"		/* to edit file */
+#define	PAGER		"more%K"	/* to view file */
+#define	EDITOR		"vi"		/* to edit file */
 
 
 /****************************************************************
@@ -48,6 +48,9 @@
 #define	HISTSIZE	50
 #define	SAVEHIST	50
 #define	DIRCOUNTLIMIT	50
+#define	SECOND		0
+#define	DOSDRIVE	0
+#define	EDITMODE	"emacs"
 #define	TMPDIR		"/tmp"
 
 
