@@ -15,6 +15,9 @@
 #include <sys/param.h>
 #include <sys/stat.h>
 
+extern int preparedrv();
+extern int shutdrv();
+
 extern int filepos;
 extern int mark;
 extern int dispmode;
@@ -22,6 +25,8 @@ extern int sorton;
 extern char *destpath;
 extern int copypolicy;
 extern char fullpath[];
+extern char *deftmpdir;
+extern char *tmpfilename;
 
 #define	realdirsiz(name)		(((strlen(name) + boundary)\
 					& ~(boundary - 1)) + dirsize)
@@ -35,15 +40,17 @@ static char *maketmpfile();
 static char *getentnum();
 static VOID restorefile();
 
+static int distdrive = -1;
 
-VOID getstatus(list, i, file)
+
+int getstatus(list, i, file)
 namelist *list;
 int i;
 char *file;
 {
 	struct stat status, lstatus;
 
-	if (lstat(file, &lstatus) < 0) error(file);
+	if (Xlstat(file, &lstatus) < 0) return(-1);
 	if (stat2(file, &status) < 0) error(file);
 	list[i].st_mode = lstatus.st_mode;
 	list[i].flags = 0;
@@ -59,6 +66,7 @@ char *file;
 	list[i].st_gid = lstatus.st_gid;
 	list[i].st_size = isdev(&list[i]) ? lstatus.st_rdev : lstatus.st_size;
 	list[i].st_mtim = lstatus.st_mtime;
+	return(0);
 }
 
 int cmplist(listp1, listp2)
@@ -71,14 +79,12 @@ namelist *listp1, *listp2;
 	if (!isdir(listp1) && isdir(listp2)) return(1);
 	if (isdir(listp1) && !isdir(listp2)) return(-1);
 
-	if (listp1 -> name[0] == '.'
-		&& (listp1 -> name[1] == '\0'
-		|| (listp1 -> name[1] == '.' && listp1 -> name[2] == '\0')))
-			return(-1);
-	if (listp2 -> name[0] == '.'
-		&& (listp2 -> name[1] == '\0'
-		|| (listp2 -> name[1] == '.' && listp2 -> name[2] == '\0')))
-			return(1);
+	if (listp1 -> name[0] == '.' && listp1 -> name[1] == '\0') return(-1);
+	if (listp2 -> name[0] == '.' && listp2 -> name[1] == '\0') return(1);
+	if (listp1 -> name[0] == '.' && listp1 -> name[1] == '.'
+	&& listp1 -> name[2] == '\0') return(-1);
+	if (listp2 -> name[0] == '.' && listp2 -> name[1] == '.'
+	&& listp2 -> name[2] == '\0') return(1);
 
 	switch (sorton & 7) {
 		case 1:
@@ -144,8 +150,8 @@ reg_t *regexp;
 {
 	struct dirent *dp;
 
-	if (!regexp) dp = readdir(dirp);
-	else while (dp = readdir(dirp)) {
+	if (!regexp) dp = Xreaddir(dirp);
+	else while (dp = Xreaddir(dirp)) {
 		if (regexp_exec(regexp, dp -> d_name)) break;
 	}
 	return(dp);
@@ -158,10 +164,11 @@ char *path;
 	int i;
 
 	cwd = getwd2();
-	if (chdir(path) < 0) path = NULL;
+	if (dospath(path, NULL) != dospath(cwd, NULL)) return(0);
+	if (_chdir2(path) < 0) path = NULL;
 	else {
 		path = getwd2();
-		if (chdir(cwd) < 0) error(cwd);
+		if (_chdir2(cwd) < 0) error(cwd);
 	}
 	i = (path) ? !strcmp(cwd, path) : 0;
 	free(cwd);
@@ -180,12 +187,12 @@ int underhome()
 			if (pwd = getpwuid(getuid())) homedir = pwd -> pw_dir;
 			else return(-1);
 		}
-		if (chdir(homedir) < 0) {
+		if (dospath(homedir, NULL) || _chdir2(homedir) < 0) {
 			homedir = NULL;
 			return(-1);
 		}
 		homedir = getwd2();
-		if (chdir(fullpath) < 0) error(fullpath);
+		if (_chdir2(fullpath) < 0) error(fullpath);
 	}
 	cp = getwd2();
 	i = strncmp(cp, homedir, strlen(homedir));
@@ -198,17 +205,26 @@ char *mes;
 {
 	struct stat status;
 	char *dir;
+	int drive;
 
-	if (!(dir = inputstr2(mes, -1, NULL, NULL))) return(NULL);
+	if (!(dir = inputstr(mes, 1, -1, NULL, NULL))) return(NULL);
 	if (!*(dir = evalpath(dir))) {
 		free(dir);
 		dir = strdup2(".");
 	}
 
+	distdrive = -1;
+	if ((drive = dospath(dir, NULL))
+	&& (distdrive = preparedrv(drive, waitmes)) < 0) {
+		warning(-1, dir);
+		free(dir);
+		return(NULL);
+	}
 	if (stat2(dir, &status) < 0) {
-		if (errno != ENOENT || mkdir(dir, 0777) < 0) {
+		if (errno != ENOENT || Xmkdir(dir, 0777) < 0) {
 			warning(-1, dir);
 			free(dir);
+			if (distdrive >= 0) shutdrv(distdrive);
 			return(NULL);
 		}
 		if (stat2(dir, &status) < 0) error(dir);
@@ -216,6 +232,7 @@ char *mes;
 	if ((status.st_mode & S_IFMT) != S_IFDIR) {
 		warning(ENOTDIR, dir);
 		free(dir);
+		if (distdrive >= 0) shutdrv(distdrive);
 		return(NULL);
 	}
 	return(dir);
@@ -230,8 +247,8 @@ int max, tr;
 		putterm(t_bell);
 		return(0);
 	}
-	destpath = (tr) ? tree(1) : getdistdir(COPYD_K);
-	if (!destpath) return((tr) ? 3 : 1);
+	destpath = (tr) ? tree(1, &distdrive) : getdistdir(COPYD_K);
+	if (!destpath) return((tr) ? 2 : 1);
 	copypolicy = (iscurdir(destpath)) ? 2 : 0;
 	if (mark > 0) applyfile(list, max, cpfile, ENDCP_K);
 	else if (isdir(&list[filepos]) && !islink(&list[filepos]))
@@ -239,6 +256,7 @@ int max, tr;
 	else if (cpfile(list[filepos].name) < 0)
 		warning(-1, list[filepos].name);
 	free(destpath);
+	if (distdrive >= 0) shutdrv(distdrive);
 	return(4);
 }
 
@@ -246,8 +264,13 @@ int movefile(list, max, tr)
 namelist *list;
 int max, tr;
 {
-	destpath = (tr) ? tree(1) : getdistdir(MOVED_K);
-	if (!destpath || iscurdir(destpath)) return((tr) ? 3 : 1);
+	if (!mark && (!strcmp(list[filepos].name, ".")
+	|| !strcmp(list[filepos].name, ".."))) {
+		putterm(t_bell);
+		return(0);
+	}
+	destpath = (tr) ? tree(1, &distdrive) : getdistdir(MOVED_K);
+	if (!destpath || iscurdir(destpath)) return((tr) ? 2 : 1);
 	copypolicy = 0;
 	if (mark > 0) filepos = applyfile(list, max, mvfile, ENDMV_K);
 	else if (mvfile(list[filepos].name) < 0)
@@ -255,9 +278,71 @@ int max, tr;
 	else filepos++;
 	if (filepos >= max) filepos = max - 1;
 	free(destpath);
+	if (distdrive >= 0) shutdrv(distdrive);
 	return(4);
 }
 
+int mktmpdir(dir)
+char *dir;
+{
+	char *cp, path[MAXPATHLEN + 1];
+	int no;
+
+	strcpy(path, deftmpdir);
+	strcat(path, "/");
+	strcat(path, tmpfilename);
+	if (mkdir(path, 0777) < 0 && errno != EEXIST) return(-1);
+	*(cp = path + strlen(path)) = '/';
+	strcpy(cp + 1, dir);
+	if (mkdir(path, 0777) < 0) {
+		*cp = '\0';
+		no = errno;
+		if (rmdir(path) < 0 && errno != ENOTEMPTY && errno != EEXIST)
+			error(path);
+		errno = no;
+		return(-1);
+	}
+	strcpy(dir, path);
+	return(0);
+}
+
+int rmtmpdir(dir)
+char *dir;
+{
+	char path[MAXPATHLEN + 1];
+
+	if (dir && *dir && rmdir(dir) < 0) return(-1);
+	strcpy(path, deftmpdir);
+	strcat(path, "/");
+	strcat(path, tmpfilename);
+	if (rmdir(path) < 0 && errno != ENOTEMPTY && errno != EEXIST)
+		return(-1);
+	return(0);
+}
+
+int forcecleandir(dir, file)
+char *dir, *file;
+{
+	extern char **environ;
+	char buf[MAXPATHLEN + 1];
+	int i, len, pid;
+
+	for (i = 0; dir[i]; i++) buf[i] = dir[i];
+	len = i;
+	buf[len++] = '/';
+	for (i = 0; file[i]; i++) buf[len + i] = file[i];
+	len += i;
+	buf[len] = '\0';
+
+	chdir("/");
+	if ((pid = fork()) < 0) return(-1);
+	else if (!pid) {
+		execle("/bin/rm", "rm", "-rf", buf, NULL, environ);
+		_exit(127);
+	}
+	return(0);
+}
+
 static VOID touch(file)
 char *file;
 {
@@ -395,10 +480,7 @@ int max, fs;
 	strcpy(path, tmpdir);
 	strcat(path, "/");
 	fnamp = strlen(path);
-	locate(0, LMESLINE);
-	putterm(l_clear);
-	kanjiputs(WAIT_K);
-	tflush();
+	waitmes();
 
 	if (!(dirp = opendir("."))) error(".");
 	i = ent = 0;
