@@ -8,6 +8,10 @@
 extern int errno;
 #endif
 
+#if !defined (ENOTEMPTY) && defined (ENFSNOTEMPTY)
+#define	ENOTEMPTY	ENFSNOTEMPTY
+#endif
+
 #ifdef	USEDIRECT
 #include <sys/dir.h>
 #else
@@ -24,15 +28,62 @@ extern int printmacro();
 extern int printlaunch();
 extern int printarch();
 extern int printalias();
+extern int printdrive();
 extern int printhist();
 extern VOID evalenv();
 
+/* dosemu.c */
+extern int _dospath();
+extern int dospath();
+extern DIR *Xopendir();
+extern int Xclosedir();
+extern struct dirent *Xreaddir();
+extern int Xchdir();
+#ifdef	USEGETWD
+extern char *Xgetwd();
+#else
+extern char *Xgetcwd();
+#endif
+extern int Xstat();
+extern int Xlstat();
+extern int Xaccess();
+extern int Xsymlink();
+extern int Xreadlink();
+extern int Xchmod();
+#ifdef	USEUTIME
+extern int Xutime();
+#else
+extern int Xutimes();
+#endif
+extern int Xunlink();
+extern int Xrename();
+extern int Xopen();
+extern int Xclose();
+extern int Xread();
+extern int Xwrite();
+extern int Xlseek();
+extern int Xmkdir();
+extern int Xrmdir();
+extern FILE *Xfopen();
+extern int Xfclose();
+extern int Xfread();
+extern int Xfwrite();
+extern int Xfflush();
+extern int Xfgetc();
+extern int Xfputc();
+extern char *Xfgets();
+extern int Xfputs();
+extern char *tmpdosdupl();
+extern int tmpdosrestore();
+
 /* libc.c */
 extern int access2();
 extern int unlink2();
 extern int rmdir2();
 extern int rename2();
 extern int stat2();
+extern char *realpath2();
+extern int _chdir2();
 extern int chdir2();
 extern char *chdir3();
 extern int mkdir2();
@@ -44,7 +95,7 @@ extern int toupper2();
 extern char *strchr2();
 extern char *strrchr2();
 extern char *strncpy2();
-extern char *strncpy3();
+extern int strncpy3();
 #ifdef	NOSTRSTR
 extern char *strstr();
 #endif
@@ -63,7 +114,8 @@ extern char *getgrgid2();
 extern time_t timelocal2();
 
 /* input.c */
-extern char *inputstr2();
+extern int getkey2();
+extern char *inputstr();
 extern int yesno();
 extern VOID warning();
 extern int selectstr();
@@ -88,14 +140,19 @@ extern int writablefs();
 extern int infofs();
 
 /* kanji.c */
+extern int getlang();
 extern int onkanji1();
+extern int jis7();
+extern int sjis2ujis();
+extern int ujis2sjis();
 extern char *mesconv();
+extern int kanjiconv();
 extern int kanjiputs();
 extern int kanjiprintf();
 extern int kanjiputs2();
 
 /* file.c */
-extern VOID getstatus();
+extern int getstatus();
 extern char *putmode();
 extern int cmplist();
 extern int cmptree();
@@ -103,9 +160,13 @@ extern struct dirent *searchdir();
 extern int underhome();
 extern int copyfile();
 extern int movefile();
+extern int mktmpdir();
+extern int rmtmpdir();
+extern int forcecleandir();
 extern VOID arrangedir();
 
 /* apply.c */
+extern int _cpfile();
 extern int cpfile();
 extern int mvfile();
 extern int cpdir();
@@ -133,7 +194,10 @@ extern char *tree();
 /* browse.c */
 extern VOID helpbar();
 extern VOID statusbar();
+extern VOID infobar();
+extern VOID waitmes();
 extern int calcwidth();
+extern VOID putname();
 extern int listupfile();
 extern VOID rewritefile();
 extern VOID movepos();
