@@ -171,6 +171,9 @@ extern char fullpath[];
 #ifndef	MNTTYPE_DGUX
 #define	MNTTYPE_DGUX	"dg/ux"	/* DG/UX */
 #endif
+#ifndef	MNTTYPE_PC
+#define	MNTTYPE_PC	"pc"	/* MS-DOS */
+#endif
 
 static int code2str();
 static int checkline();
@@ -216,6 +219,12 @@ int code;
 			case K_BS:
 				cp = "Bs";
 				break;
+			case K_DL:
+				cp = "DelLin";
+				break;
+			case K_IL:
+				cp = "InsLin";
+				break;
 			case K_DC:
 				cp = "Del";
 				break;
@@ -225,12 +234,24 @@ int code;
 			case K_CLR:
 				cp = "Clr";
 				break;
+			case K_EOL:
+				cp = "Eol";
+				break;
 			case K_NPAGE:
 				cp = "PageDn";
 				break;
 			case K_PPAGE:
 				cp = "PageUp";
 				break;
+			case K_ENTER:
+				cp = "Enter";
+				break;
+			case K_BEG:
+				cp = "Beg";
+				break;
+			case K_END:
+				cp = "End";
+				break;
 			default:
 				return(0);
 		}
@@ -271,13 +292,13 @@ int mode;
 
 		c = 0;
 		*buf = '\0';
-		for (j = 0; j < MAXBINDTABLE && bindlist[j].key > 0; j++)
+		for (j = 0; j < MAXBINDTABLE && bindlist[j].key >= 0; j++)
 			if (i == (int)(bindlist[j].f_func)) {
 				if ((c += code2str(buf,
 					(int)(bindlist[j].key))) >= 2) break;
 			}
 		if (c < 2)
-		for (j = 0; j < MAXBINDTABLE && bindlist[j].key > 0; j++)
+		for (j = 0; j < MAXBINDTABLE && bindlist[j].key >= 0; j++)
 			if (i == (int)(bindlist[j].d_func)) {
 				if ((c += code2str(buf,
 					(int)(bindlist[j].key))) >= 2) break;
@@ -350,7 +371,7 @@ mnt_t *mntp;
 	dir = (char *)realloc2(dir, len);
 	strcpy(dir, cp);
 
-	entp = (char *)getvfsbytype(vmntp -> vmt_gfstype);
+	entp = getvfsbytype(vmntp -> vmt_gfstype);
 	if (entp) {
 		cp = entp -> vfsent_name;
 		len = strlen(cp) + 1;
@@ -364,7 +385,7 @@ mnt_t *mntp;
 
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
-	mntp -> mnt_type = type;
+	mntp -> mnt_type = (type) ? type : "???";
 	mntp -> mnt_opts = (opts) ? opts : "";
 
 	mnt_ptr += vmntp -> vmt_length;
@@ -397,7 +418,7 @@ char *file, *mode;
 	return((FILE *)buf);
 }
 
-#ifdef	USEMNTINFOR
+#if !defined(MNT_RDONLY) && defined(M_RDONLY)
 #define	MNT_RDONLY	M_RDONLY
 #endif
 
@@ -410,16 +431,24 @@ mnt_t *mntp;
 	static char *type = NULL;
 	struct statfs *buf;
 #ifdef	USEMNTINFO
-# ifdef	INITMOUNTNAMES
-	static char *mnt_names[] = INITMOUNTNAMES;
-# define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) ? \
-					mnt_names[n] : NULL)
-# else
+# ifdef	USEVFCNAME
 	struct vfsconf *conf;
 # define	getvfsbynumber(n)	((conf = getvfsbytype(n)) ? \
 					conf -> vfc_name : NULL)
+# else
+#  ifdef	USEFFSTYPE
+#  define	getvfsbynumber(n)	(buf[mnt_ptr].f_fstypename)
+#  else
+#   ifdef	INITMOUNTNAMES
+	static char *mnt_names[] = INITMOUNTNAMES;
+#   define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) ? \
+					mnt_names[n] : NULL)
+#   else
+#   define	getvfsbynumber(n)	(NULL)
+#   endif
+#  endif
 # endif
-#else
+#else	/* USEMNTINFO */
 # ifdef	USEGETFSSTAT
 # define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) ? \
 					mnt_names[n] : NULL)
@@ -439,7 +468,7 @@ mnt_t *mntp;
 	dir = (char *)realloc2(dir, len);
 	strcpy(dir, buf[mnt_ptr].f_mntonname);
 
-	cp = getvfsbynumber(buf[mnt_ptr].f_type);
+	cp = (char *)getvfsbynumber(buf[mnt_ptr].f_type);
 	if (cp) {
 		len = strlen(cp) + 1;
 		type = (char *)realloc2(type, len);
@@ -452,7 +481,7 @@ mnt_t *mntp;
 
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
-	mntp -> mnt_type = type;
+	mntp -> mnt_type = (type) ? type : "???";
 	mntp -> mnt_opts = (buf[mnt_ptr].f_flags & MNT_RDONLY) ? "ro" : "";
 
 	mnt_ptr++;
@@ -517,13 +546,13 @@ mnt_t *mntbuf;
 
 	dir = NULL;
 	if (!strncmp(path, "/dev/", 4)) {
-		if (chdir(path) < 0) dir = path;
+		if (_chdir2(path) < 0) dir = path;
 	}
-	else if (chdir(path) < 0) return(0);
+	else if (dospath(path, NULL) || _chdir2(path) < 0) return(0);
 
 	if (!dir) {
 		dir = getwd2();
-		if (chdir(fullpath) < 0) error(fullpath);
+		if (_chdir2(fullpath) < 0) error(fullpath);
 		match = 0;
 
 		fp = setmntent(MOUNTED, "r");
@@ -560,6 +589,7 @@ char *path;
 	statfs_t fsbuf;
 	mnt_t mntbuf;
 
+	if (dospath(path, NULL)) return(0);
 	if (access(path, R_OK | W_OK | X_OK) < 0) return(-1);
 	if (!getfsinfo(path, &fsbuf, &mntbuf)
 	|| hasmntopt(&mntbuf, "ro")) return(0);
