@@ -12,47 +12,47 @@
 #define	KANA	1
 #define	KANJI	2
 
-#ifdef	CPP7BIT
-static int fputs2();
-static int through;
-#else
-#define	fputs2	fputs
-#endif
+#define	SJIS	3
+#define	EUC	4
 
+static int fputs2();
 static unsigned char *convert();
 static int output();
 
+static int msboff = 0;
+static int prefix = 0;
+static int removebs = 0;
+static int kanjicode = SJIS;
+
 
-#ifdef	CPP7BIT
 static int fputs2(s, fp)
 char *s;
 FILE *fp;
 {
 	int c;
 
-	if (through) return(fputs(s, fp));
+	if (!msboff) return(fputs(s, fp));
 	while (c = *(unsigned char *)(s++)) {
 		if (c & 0x80) fprintf(fp, "\\%o", c);
 		else fputc(c, fp);
 	}
 	return(0);
 }
-#endif
 
 static unsigned char *convert(j1, j2)
 int j1, j2;
 {
 	static unsigned char cnv[4];
 
-#ifdef	CODEEUC
-	cnv[0] = (j1) ? j1 | 0x80 : 0x8e;
-	cnv[1] = j2 | 0x80;
-	cnv[2] = '\0';
-#else
-	if (j1) {
+	if (kanjicode == EUC) {
+		cnv[0] = (j1) ? j1 | 0x80 : 0x8e;
+		cnv[1] = j2 | 0x80;
+		cnv[2] = '\0';
+	}
+	else if (j1) {
 		cnv[0] = ((j1 - 1) >> 1) + ((j1 < 0x5e) ? 0x71 : 0xb1);
 		cnv[1] = j2 + ((j1 & 1) ? ((j2 < 0x60) ? 0x1f : 0x20) : 0x7e);
-		if (cnv[1] == '\\') {
+		if (cnv[1] == '\\' && prefix) {
 			cnv[2] = '\\';
 			cnv[3] = '\0';
 		}
@@ -62,7 +62,6 @@ int j1, j2;
 		cnv[0] = j2 | 0x80;
 		cnv[1] = '\0';
 	}
-#endif
 	return(cnv);
 }
 
@@ -71,25 +70,26 @@ FILE *fp;
 int c;
 int mode;
 {
+	static unsigned char buf[2];
 	static int kanji1 = 0;
+	static int bufp = 0;
 
-	switch (mode) {
-		case ASCII:
-			fputc(c, fp);
-			kanji1 = 0;
-			break;
-		case KANJI:
-			if (kanji1) {
-				fputs2((char *)convert(kanji1, c), fp);
-				kanji1 = 0;
-			}
-			else kanji1 = c;
-			break;
-		default:
-			fputs2((char *)convert(0, c), fp);
-			kanji1 = 0;
-			break;
+	if (!fp) bufp = kanji1 = 0;
+
+	if (bufp > 1) {
+		fputs2((char *)convert(buf[0], buf[1]), fp);
+		bufp = 0;
+		kanji1 = mode;
 	}
+	else if (bufp > 0 && !(kanji1 & KANJI)) {
+		if (kanji1 & KANA) fputs2((char *)convert(0, buf[0]), fp);
+		else fputc(buf[0], fp);
+		bufp = 0;
+		kanji1 = mode;
+	}
+	else if (!bufp) kanji1 = mode;
+
+	if (c != EOF) buf[bufp++] = c;
 }
 
 int main(argc, argv)
@@ -97,16 +97,42 @@ int argc;
 char *argv[];
 {
 	FILE *in, *out;
-	int c, mode, esc, kanji;
+	int i, j, c, mode, esc, kanji;
 
-#ifdef	CPP7BIT
-	char *cp;
+	for (i = 1; i < argc; i++) {
+		if (argv[i][0] != '-' || !argv[i][1]) break;
+		if (argv[i][1] == '-' && !argv[i][2]) {
+			i++;
+			break;
+		}
+		for (j = 1; argv[i][j]; j++) switch (argv[i][j]) {
+			case '7':
+				msboff = 1;
+				break;
+			case 'e':
+				kanjicode = EUC;
+				break;
+			case 's':
+				kanjicode = SJIS;
+				break;
+			case 'c':
+				prefix = 1;
+				break;
+			case 'b':
+				removebs = 1;
+				break;
+			default:
+				break;
+		}
+	}
+	if (i >= argc) {
+		fprintf(stderr,
+			"\007Usage: kanjicnv [-7bces] <infile> [<outfile>]\n");
+		exit(1);
+	}
 
-	through = ((cp = strrchr(argv[2], '.')) && !strcmp(++cp, "h")) ? 0 : 1;
-#endif
-
-	in = fopen(argv[1], "r");
-	out = fopen(argv[2], "w");
+	in = strcmp(argv[i], "-") ? fopen(argv[i], "r") : stdin;
+	out = (i + 1 < argc) ? fopen(argv[i + 1], "w") : stdout;
 
 	mode = ASCII;
 	esc = kanji = 0;
@@ -136,6 +162,11 @@ char *argv[];
 			case '\017':	/* SO */
 				mode &= ~KANA;
 				break;
+			case '\b':
+				if (removebs) {
+					output(NULL, EOF, mode);
+					break;
+				}
 			default:
 				if (esc) output(out, '\033', mode);
 				if (kanji > 0) mode |= KANJI;
@@ -145,6 +176,7 @@ char *argv[];
 				break;
 		}
 	}
+	output(out, EOF, mode);
 
 	fclose(out);
 	fclose(in);
