@@ -10,6 +10,7 @@
 #include "kctype.h"
 #include "kanji.h"
 
+#include <fcntl.h>
 #include <pwd.h>
 #include <grp.h>
 #include <sys/file.h>
@@ -26,7 +27,7 @@
 #endif
 
 #ifdef	USEGETWD
-#define	getcwd(buf, size)	getwd(buf)
+#define	Xgetcwd(buf, size)	Xgetwd(buf)
 #endif
 
 #ifdef	PWNEEDERROR
@@ -35,16 +36,18 @@ char Error[1024];
 
 extern int copypolicy;
 extern char fullpath[];
-extern char *lastpath;
 extern char *origpath;
 extern char *findpattern;
 
 #define	BUFUNIT		32
 
+static char *_realpath2();
 static assoclist *_getenv2();
 static long gettimezone();
 
 static assoclist *environ2 = NULL;
+static char *lastpath = NULL;
+static char *unixpath = NULL;
 
 
 int access2(path, mode)
@@ -55,6 +58,7 @@ int mode;
 	char *cp, *name, dir[MAXPATHLEN + 1], *str[4];
 	int val[4];
 
+	if (dospath(path, NULL)) return(Xaccess(path, mode));
 	if (lstat(path, &status) < 0) error(path);
 	if ((status.st_mode & S_IFMT) == S_IFLNK) return(0);
 
@@ -100,21 +104,21 @@ int unlink2(path)
 char *path;
 {
 	if (access2(path, W_OK) < 0) return(1);
-	return(unlink(path));
+	return(Xunlink(path));
 }
 
 int rmdir2(path)
 char *path;
 {
 	if (access2(path, R_OK | W_OK | X_OK) < 0) return(1);
-	return(rmdir(path));
+	return(Xrmdir(path));
 }
 
 int rename2(from, to)
 char *from, *to;
 {
 	if (!strcmp(from, to)) return(0);
-	return(rename(from, to));
+	return(Xrename(from, to));
 }
 
 int stat2(path, buf)
@@ -123,7 +127,8 @@ struct stat *buf;
 {
 	int tmperr;
 
-	if (stat(path, buf) < 0) {
+	if (Xstat(path, buf) < 0) {
+		if (dospath(path, NULL)) return(-1);
 		tmperr = errno;
 		if (lstat(path, buf) < 0
 		|| (buf -> st_mode & S_IFMT) != S_IFLNK) {
@@ -135,29 +140,75 @@ struct stat *buf;
 	return(0);
 }
 
-static VOID _chdir2(path)
-char *path;
+static char *_realpath2(path, resolved)
+char *path, *resolved;
 {
 	char *cp;
+	int drive;
 
-	if (!*path || !strcmp(path, ".")) return;
+	if (!*path || !strcmp(path, ".")) return(resolved);
 	else if (cp = strchr(path, '/')) {
-		*cp = 0;
-		_chdir2(path);
+		*cp = '\0';
+		_realpath2(path, resolved);
 		*(cp++) = '/';
-		_chdir2(cp);
-		return;
+		_realpath2(cp, resolved);
+		return(resolved);
 	}
 
 	if (!strcmp(path, "..")) {
-		cp = strrchr(fullpath, '/');
-		if (cp && cp != fullpath) *cp = '\0';
-		else strcpy(fullpath, "/");
+		cp = strrchr(resolved, '/');
+		if (drive = _dospath(resolved)) {
+			if (cp && cp != &resolved[2]) *cp = '\0';
+			else sprintf(resolved, "%c:/", toupper2(drive));
+		}
+		else {
+			if (cp && cp != resolved) *cp = '\0';
+			else strcpy(resolved, "/");
+		}
 	}
 	else {
-		if (strcmp(fullpath, "/")) strcat(fullpath, "/");
-		strcat(fullpath, path);
+		if (*resolved && resolved[strlen(resolved) - 1] != '/')
+			strcat(resolved, "/");
+		strcat(resolved, path);
 	}
+	return(resolved);
+}
+
+char *realpath2(path, resolved)
+char *path, *resolved;
+{
+	char tmp[MAXPATHLEN + 1];
+
+	if (path == resolved) {
+		strcpy(tmp, path);
+		path = tmp;
+	}
+
+	if (*path == '/') strcpy(resolved, "/");
+	else if (_dospath(path)) {
+		sprintf(resolved, "%c:/", toupper2(*path));
+		path += 2;
+	}
+	else if (resolved != fullpath) strcpy(resolved, fullpath);
+	return(_realpath2(path, resolved));
+}
+
+int _chdir2(path)
+char *path;
+{
+	char cwd[MAXPATHLEN + 1];
+	int fd;
+
+	if (!Xgetcwd(cwd, MAXPATHLEN)) strcpy(cwd, "/");
+	if (Xchdir(path) < 0) return(-1);
+	if (!dospath("", NULL)) {
+		if ((fd = open(".", O_RDONLY, 0600)) < 0) {
+			if (Xchdir(cwd) < 0) error(cwd);
+			return(-1);
+		}
+		close(fd);
+	}
+	return(0);
 }
 
 int chdir2(path)
@@ -168,19 +219,31 @@ char *path;
 #endif
 	char cwd[MAXPATHLEN + 1];
 
-	if (access(path, R_OK | X_OK) < 0 || chdir(path) < 0) return(-1);
+	if (_chdir2(path) < 0) return(-1);
 
 	strcpy(cwd, fullpath);
-	if (*path == '/') strcpy(fullpath, "/");
-	_chdir2(path);
+	realpath2(path, fullpath);
 
-	if (chdir(fullpath) < 0) {
-		if (chdir(cwd) < 0) error(cwd);
+	if (_chdir2(fullpath) < 0) {
+		if (_chdir2(cwd) < 0) error(cwd);
 		strcpy(fullpath, cwd);
 		return(-1);
 	}
 	if (lastpath) free(lastpath);
 	lastpath = strdup2(cwd);
+	if (_dospath(fullpath)) {
+		if (!_dospath(cwd)) {
+			if (unixpath) free(unixpath);
+			unixpath = strdup2(cwd);
+		}
+		if (Xgetcwd(cwd, MAXPATHLEN)) strcpy(fullpath, cwd);
+		realpath2(fullpath, fullpath);
+	}
+	else {
+		if (unixpath) free(unixpath);
+		unixpath = NULL;
+	}
+
 	if (getenv("PWD")) {
 #ifdef	USESETENV
 		setenv("PWD", fullpath, 1);
@@ -205,6 +268,10 @@ char *path;
 		if (!lastpath) return(".");
 		path = lastpath;
 	}
+	else if (!strcmp(path, "@")) {
+		if (!unixpath) return(".");
+		path = unixpath;
+	}
 	if (chdir2(path) < 0) return(NULL);
 	if (!cwd) {
 		cwd = getwd2();
@@ -232,7 +299,7 @@ int mode;
 	cp1 = ++eol;
 	cp2 = strrchr(path, '/');
 	for (;;) {
-		if (mkdir(path, mode) >= 0) break;
+		if (Xmkdir(path, mode) >= 0) break;
 		if (errno != ENOENT || !cp2 || cp2 <= path) return(-1);
 		*cp2 = '\0';
 		if (cp1 < eol) *cp1 = '/';
@@ -244,7 +311,7 @@ int mode;
 		cp2 = strchr(cp1 + 1, '/');
 		*cp1 = '/';
 		if (cp2) *cp2 = '\0';
-		if (mkdir(path, mode) < 0 && errno != EEXIST) return(-1);
+		if (Xmkdir(path, mode) < 0 && errno != EEXIST) return(-1);
 		cp1 = cp2;
 	}
 	return(0);
@@ -341,29 +408,32 @@ int n;
 	return(s1);
 }
 
-char *strncpy3(s1, s2, len, ptr)
+int strncpy3(s1, s2, len, ptr)
 char *s1, *s2;
 int len, ptr;
 {
 	char *cp;
-	int i;
+	int i, l;
 
 	cp = s1;
+	l = len;
 	if (ptr && onkanji1((u_char *)s2, ptr - 1)) {
 		*(cp++) = ' ';
 		ptr++;
+		l--;
+	}
+	if (l < strlen(s2 + ptr) && onkanji1((u_char *)s2, ptr + l - 1)) {
+		strncpy2(cp, s2 + ptr, --l);
+		strcat(cp, " ");
 		len--;
 	}
-	if (len < strlen(s2 + ptr) && onkanji1((u_char *)s2, ptr + len - 1)) {
-		strncpy2(cp, s2 + ptr, --len);
-		strcat(cp, " ");
-	}
 	else {
-		for (i = 0; i < len && s2[ptr + i]; i++) *(cp++) = s2[ptr + i];
-		for (; i < len; i++) *(cp++) = ' ';
+		for (i = 0; i < l && s2[ptr + i]; i++) *(cp++) = s2[ptr + i];
+		len = i;
+		for (; i < l; i++) *(cp++) = ' ';
 		*cp = '\0';
 	}
-	return(s1);
+	return(len);
 }
 
 #ifdef	NOSTRSTR
@@ -439,6 +509,7 @@ int overwrite;
 		free(ap -> assoc);
 	}
 	else {
+		if (!value) return(0);
 		ap = (assoclist *)malloc2(sizeof(assoclist));
 		ap -> org = strdup2(name);
 		ap -> next = environ2;
@@ -515,10 +586,12 @@ int noconf;
 	cooked2();
 	echo2();
 	nl2();
+	tabs();
 	status = system2(command);
 	raw2();
 	noecho2();
 	nonl2();
+	notabs();
 	if (status > 127 || !noconf) warning(0, HITKY_K);
 	if (noconf >= 0) {
 		if (noconf) putterms(t_init);
@@ -532,7 +605,7 @@ char *getwd2()
 {
 	char cwd[MAXPATHLEN + 1];
 
-	if (!getcwd(cwd, MAXPATHLEN)) error(NULL);
+	if (!Xgetcwd(cwd, MAXPATHLEN)) error(NULL);
 	return(strdup2(cwd));
 }
 
@@ -595,18 +668,27 @@ static long gettimezone()
 #ifdef	HAVETIMEZONE
 	extern time_t timezone;
 
+	tzset();
 	tz = timezone;
 #else
+# ifdef	NOTMGMTOFF
 	struct timeval t_val;
 	struct timezone t_zone;
 
 	gettimeofday(&t_val, &t_zone);
 	tz = t_zone.tz_minuteswest * 60L;
+# else
+	tz = time(0);
+	tz = -(localtime(&tz) -> tm_gmtoff);
+# endif
 #endif
 #ifdef	USELEAPCNT
-	strcpy(path, TZDIR);
-	strcat(path, "/");
-	strcat(path, TZDEFAULT);
+	if (TZDEFAULT[0] == '/') strcpy(path, TZDEFAULT);
+	else {
+		strcpy(path, TZDIR);
+		strcat(path, "/");
+		strcat(path, TZDEFAULT);
+	}
 	if (fp = fopen(path, "r")) {
 		if (fread(&buf, sizeof(struct tzhead), 1, fp) == 1)
 			tz += *((long *)(buf.tzh_leapcnt));
