@@ -14,6 +14,7 @@
 #  define	USERE_COMP
 #  endif
 # define	REGEXPLIB	"-lgen"
+# define	NODNAMLEN
 # define	USESTATVFSH
 # define	USEMNTTABH
 # else
@@ -29,6 +30,7 @@
 # define	SVR4
 # define	OSTYPE		"NEWS_OS6"
 # define	REGEXPLIB	"-lgen"
+# define	NODNAMLEN
 # define	USESTATVFSH
 # define	USEMNTTABH
 # define	USEUTIME
@@ -51,6 +53,7 @@
 #   define	NOUNISTDH
 #   define	NOSTDLIBH
 #   define	USEDIRECT
+#   define	NOTMGMTOFF
 #   define	NOVSPRINTF
 #   define	USEGETWD
 #   endif
@@ -69,6 +72,7 @@
 # define	TERMCAPLIB	"-lcurses"
 # define	EXTENDLIB	"-lsun"
 # endif
+#define	NODNAMLEN
 #define	USESTATFSH
 #define	STATFSARGS	4
 #define	USERE_COMP
@@ -96,6 +100,7 @@
 #  endif
 # define	TERMCAPLIB	"-lcurses"
 # define	REGEXPLIB	"-lgen"
+# define	NODNAMLEN
 # define	USESTATVFSH
 # define	USEMNTTABH
 # define	USEUTIME
@@ -119,6 +124,7 @@
 #define	USETIMEH
 #define	USETERMIO
 #define	HAVETIMEZONE
+#define	SYSVDIRENT
 #define	USEREGCMP
 #define	NOSTRSTR
 #define	WAITKEYPAD		720
@@ -170,6 +176,7 @@
 # define	SVR4
 # define	OSTYPE		"DECOSF1V3"
 # define	EXTENDLIB	"-lc_r"
+# define	NODNAMLEN
 # define	USESTATVFSH
 # define	USEMNTINFOR
 # define	USEREGCOMP
@@ -219,6 +226,9 @@
 #define	TERMCAPLIB		"-ltermcap"
 #define	USESTATFSH
 #define	STATFSARGS	4
+#define	USESYSDIRH
+#define	NODNAMLEN
+#define	USERE_COMP
 #endif
 
 #if defined (__uxpm__)
@@ -226,9 +236,10 @@
 #define	OSTYPE			"UXPM"
 #define	CODEEUC
 #define	REGEXPLIB		"-lgen"
+#define	NODNAMLEN
 #define	USESTATVFSH
 #define	USEMNTTABH
-#define	USERE_COMP
+#define	USEUTIME
 #endif
 
 #if defined (mips) && !defined (OSTYPE)
@@ -256,11 +267,25 @@
 #define	SVR4
 #define	OSTYPE			"LINUX"
 #define	CODEEUC
-#define	TERMCAPLIB		"-lncurses"
+#define	TERMCAPLIB		"-ltermcap"
 #define	USERE_COMP
 #endif
 
-#if defined (__FreeBSD__)
+#if defined (__FreeBSD__) && defined (__powerpc__)
+#define	BSD43
+#define	OSTYPE			"JCCBSD"
+#define	CODEEUC
+#define	TARUSESPACE
+#define	REGEXPLIB		"-lcompat"
+#define	USELEAPCNT
+#define	DECLERRLIST
+#define	USEMOUNTH
+#define	USEMNTINFO
+#define	USERE_COMP
+#define	USESETENV
+#endif
+
+#if defined (__FreeBSD__) && !defined (OSTYPE)
 #define	BSD43
 #define	OSTYPE			"FREEBSD"
 #define	CODEEUC
@@ -269,6 +294,7 @@
 #define	DECLERRLIST
 #define	USEMOUNTH
 #define	USEMNTINFO
+#define	USEVFCNAME
 #define	USERE_COMP
 #define	USESETENV
 #endif
@@ -276,11 +302,19 @@
 #if defined (__NetBSD__)
 #define	BSD43
 #define	OSTYPE			"NETBSD"
+#define	CODEEUC
+#define	TARUSESPACE
+#define	REGEXPLIB		"-lcompat"
 #define	DECLERRLIST
 #define	USEMOUNTH
-#define	USEFFSIZE
 #define	USEMNTINFO
+#define	USEFFSTYPE
 #define	USERE_COMP
+#define	USESETENV
+#include <sys/param.h>
+# if (NetBSD1_0 < 1)
+# define	USEFFSIZE
+# endif
 #endif
 
 #if defined (__bsdi__)
@@ -295,11 +329,35 @@
 #define	USERE_COMP
 #define	USESETENV
 #include <sys/param.h>
-# if (BSD4_4 != 1)
+# if !defined (BSD) || (BSD < 199306)
 # define	USEFFSIZE
 # endif
 #endif
 
+#if defined (__386BSD__) && defined (__BSD_NET2) && !defined (OSTYPE)
+#define	BSD43
+#define	OSTYPE			"BOW"
+#define	TARUSESPACE
+#define	TERMCAPLIB		"-ltermcap"
+#define	DECLERRLIST
+#define	USEMOUNTH
+#define	USEMNTINFO
+#define	USEVFCNAME
+#define	USEREGCOMP
+#define	USESETENV
+#endif
+
+#if defined (__386BSD__) && !defined (OSTYPE)
+#define	BSD43
+#define	OSTYPE			"ORG_386BSD"
+#define	TARUSESPACE
+#define	DECLERRLIST
+#define	SYSVDIRENT
+#define	USEMOUNTH
+#define	USEMNTINFO
+#define	USESETENV
+#endif
+
 /****************************************************************
  *	If your machine is not described above,	you should	*
  *	comment out below manually to apply your environment.	*
@@ -332,9 +390,12 @@
 /*	UXPM		/* UXP/M (Fujitsu) */
 /*	MIPS		/* RISC/os (MIPS) */
 /*	LINUX		/* Linux */
-/*	FREEBSD		/* 4.4BSD-Lite (JCC) or FreeBSD */
+/*	JCCBSD		/* 4.4BSD-Lite (JCC) */
+/*	FREEBSD		/* FreeBSD */
 /*	NETBSD		/* NetBSD */
 /*	BSDOS		/* BSD/OS (BSDI) */
+/*	BOW		/* BSD on Windows (ASCII) */
+/*	386BSD_ORG	/* 386BSD */
 
 /* #define CODEEUC	/* kanji code type is EUC */
 /* #define TARUSESPACE	/* tar(1) uses space to devide file mode from UID */
@@ -362,7 +423,9 @@
 /* #define USETERMIO	/* use termio interface */
 /* #define USEDIRECT	/* use 'struct direct' instead of dirent */
 /* #define SYSVDIRENT	/* dirent interface behaves as System V */
+/* #define NODNAMLEN	/* struct dirent haven't d_namlen */
 /* #define HAVETIMEZONE	/* have extern valiable 'timezone' */
+/* #define NOTMGMTOFF	/* struct tm haven't tm_gmtoff */
 
 /* following 5 items are exclusive
 /* #define USESTATVFSH	/* use <sys/statvfs.h> as header of the FS status */
@@ -384,6 +447,10 @@
 /* #define USEGETMNT	/* use getmnt() to get the mount entry */
 /* #define USEGETFSENT	/* use getfsent() to get the mount entry */
 
+/* following 2 items are exclusive
+/* #define USEVFCNAME	/* 'struct vfsconf' has 'vfc_name' as the mount type */
+/* #define USEFFSTYPE	/* 'struct statfs' has 'f_fstypename' as mount type */
+
 /* following 3 items are exclusive
 /* #define USERE_COMP	/* use re_comp() family as search */
 /* #define USEREGCOMP	/* use regcomp() family as search */
@@ -423,19 +490,19 @@
 #define	REGEXPLIB		"-lPW"
 #endif
 
-#ifndef CCCOMMAND
+#ifndef	CCCOMMAND
 #define	CCCOMMAND		"cc"
 #endif
-#ifndef EXTENDCCOPT
+#ifndef	EXTENDCCOPT
 #define	EXTENDCCOPT		"-O"
 #endif
-#ifndef TERMCAPLIB
+#ifndef	TERMCAPLIB
 #define	TERMCAPLIB		"-ltermlib"
 #endif
-#ifndef REGEXPLIB
+#ifndef	REGEXPLIB
 #define	REGEXPLIB
 #endif
-#ifndef EXTENDLIB
+#ifndef	EXTENDLIB
 #define	EXTENDLIB
 #endif
 
