@@ -11,6 +11,7 @@
 #include "kanji.h"
 #include "funcno.h"
 #include "version.h"
+#include "dosdisk.h"
 
 #include <signal.h>
 #include <sys/time.h>
@@ -35,12 +36,17 @@ extern aliastable aliaslist[];
 extern int maxalias;
 extern bindtable bindlist[];
 extern functable funclist[];
+extern devinfo fdtype[];
 extern char *archivefile;
 extern char **sh_history;
 extern char *helpindex[];
 extern int subwindow;
 extern int columns;
 extern int dispmode;
+extern int dosdrive;
+extern char *editmode;
+extern int inputkcode;
+extern int outputkcode;
 
 #define	CLOCKUPDATE	10	/* sec */
 
@@ -56,6 +62,38 @@ extern int dispmode;
 # endif
 #endif
 
+#if !defined(SIGIOT) && defined(SIGABRT)
+#define	SIGIOT	SIGABRT
+#endif
+
+static VOID signalexit();
+static sigarg_t ignore_alrm();
+static sigarg_t ignore_winch();
+static sigarg_t ignore_int();
+static sigarg_t ignore_quit();
+static sigarg_t hungup();
+static sigarg_t illerror();
+static sigarg_t traperror();
+#ifdef	SIGIOT
+static sigarg_t ioerror();
+#endif
+#ifdef	SIGEMT
+static sigarg_t emuerror();
+#endif
+static sigarg_t floaterror();
+static sigarg_t buserror();
+static sigarg_t segerror();
+#ifdef	SIGSYS
+static sigarg_t syserror();
+#endif
+static sigarg_t pipeerror();
+static sigarg_t terminate();
+#ifdef	SIGXCPU
+static sigarg_t xcpuerror();
+#endif
+#ifdef	SIGXFSZ
+static sigarg_t xsizerror();
+#endif
 static sigarg_t wintr();
 static sigarg_t printtime();
 static char *skipspace();
@@ -66,6 +104,7 @@ static int getlaunch();
 static int getcommand();
 static int getkeybind();
 static int getalias();
+static int getdosdrive();
 static int unalias();
 static int loadruncom();
 static VOID printext();
@@ -78,8 +117,9 @@ int minfilename;
 int histsize;
 int savehist;
 int dircountlimit;
+int showsecond;
 char *deftmpdir;
-char *language;
+char *tmpfilename;
 
 static char *progname;
 static int timersec = 0;
@@ -91,58 +131,175 @@ char *str;
 	if (!str) str = progname;
 	endterm();
 	inittty(1);
+	fputc('\007', stderr);
 	perror(str);
-	exit2(1);
+	forcecleandir(deftmpdir, tmpfilename);
+	dosallclose();
+	exit(127);
 }
 
+static VOID signalexit(sig)
+int sig;
+{
+	signal(sig, SIG_IGN);
+	inittty(1);
+	forcecleandir(deftmpdir, tmpfilename);
+	dosallclose();
+	signal(sig, SIG_DFL);
+	kill(getpid(), sig);
+}
+
+static sigarg_t ignore_alrm()
+{
+	signal(SIGALRM, (sigarg_t (*)())ignore_alrm);
+}
+
+static sigarg_t ignore_winch()
+{
+	signal(SIGWINCH, (sigarg_t (*)())ignore_winch);
+}
+
+static sigarg_t ignore_int()
+{
+	signal(SIGINT, (sigarg_t (*)())ignore_int);
+}
+
+static sigarg_t ignore_quit()
+{
+	signal(SIGQUIT, (sigarg_t (*)())ignore_quit);
+}
+
+static sigarg_t hangup()
+{
+	signalexit(SIGHUP);
+}
+
+static sigarg_t illerror()
+{
+	signalexit(SIGILL);
+}
+
+static sigarg_t traperror()
+{
+	signalexit(SIGTRAP);
+}
+
+#ifdef	SIGIOT
+static sigarg_t ioerror()
+{
+	signalexit(SIGIOT);
+}
+#endif
+
+#ifdef	SIGEMT
+static sigarg_t emuerror()
+{
+	signalexit(SIGEMT);
+}
+#endif
+
+static sigarg_t floaterror()
+{
+	signalexit(SIGFPE);
+}
+
+static sigarg_t buserror()
+{
+	signalexit(SIGBUS);
+}
+
+static sigarg_t segerror()
+{
+	signalexit(SIGSEGV);
+}
+
+#ifdef	SIGSYS
+static sigarg_t syserror()
+{
+	signalexit(SIGSYS);
+}
+#endif
+
+static sigarg_t pipeerror()
+{
+	signalexit(SIGPIPE);
+}
+
+static sigarg_t terminate()
+{
+	signalexit(SIGTERM);
+}
+
+#ifdef	SIGXCPU
+static sigarg_t xcpuerror()
+{
+	signalexit(SIGXCPU);
+}
+#endif
+
+#ifdef	SIGXFSZ
+static sigarg_t xsizerror()
+{
+	signalexit(SIGXFSZ);
+}
+#endif
+
 static sigarg_t wintr()
 {
 	signal(SIGWINCH, SIG_IGN);
 	getwsize(80, WHEADER + WFOOTER + 2);
 	title();
-	if (archivefile) rewritearc();
-	else rewritefile();
-	if (subwindow) ungetch2(CTRL_L);
-	signal(SIGWINCH, (sigarg_t (*))wintr);
+	if (archivefile) rewritearc(1);
+	else rewritefile(1);
+	if (subwindow) ungetch2(CTRL('L'));
+	signal(SIGWINCH, (sigarg_t (*)())wintr);
 }
 
 static sigarg_t printtime()
 {
+	static long now;
 	struct timeval t;
 	struct timezone tz;
 	struct tm *tm;
 
 	signal(SIGALRM, SIG_IGN);
-	if (!timersec) {
+	if (timersec) now++;
+	else {
 		gettimeofday(&t, &tz);
-		tzset();
-		tm = localtime(&(t.tv_sec));
-
-		locate(n_column - 16, LTITLE);
+		now = t.tv_sec;
+		timersec = CLOCKUPDATE;
+	}
+	if (showsecond || timersec == CLOCKUPDATE) {
+		tm = localtime(&now);
+		locate(n_column - 16 - ((showsecond) ? 3 : 0), LTITLE);
 		putterm(t_standout);
-		cprintf("%02d-%02d-%02d %02d:%02d",
+		if (showsecond) cprintf("%02d-%02d-%02d %02d:%02d:%02d",
+			tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday,
+			tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
+		else cprintf("%02d-%02d-%02d %02d:%02d",
 			tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday,
 			tm -> tm_hour, tm -> tm_min);
 		putterm(end_standout);
 		locate(0, 0);
 		tflush();
-		timersec = CLOCKUPDATE;
 	}
 	timersec--;
-	signal(SIGALRM, (sigarg_t (*))printtime);
+	signal(SIGALRM, (sigarg_t (*)())printtime);
 }
 
 VOID sigvecset()
 {
 	getwsize(80, WHEADER + WFOOTER + 2);
-	signal(SIGWINCH, (sigarg_t (*))wintr);
-	signal(SIGALRM, (sigarg_t (*))printtime);
+	signal(SIGALRM, (sigarg_t (*)())printtime);
+	signal(SIGTSTP, SIG_IGN);
+	signal(SIGWINCH, (sigarg_t (*)())wintr);
 }
 
 VOID sigvecreset()
 {
-	signal(SIGWINCH, SIG_IGN);
-	signal(SIGALRM, SIG_IGN);
+	signal(SIGALRM, (sigarg_t (*)())ignore_alrm);
+	signal(SIGTSTP, SIG_DFL);
+	signal(SIGWINCH, (sigarg_t (*)())ignore_winch);
 }
 
 VOID title()
@@ -151,13 +308,13 @@ VOID title()
 
 	locate(0, LTITLE);
 	putterm(t_standout);
-	cputs("  FD (File & Directory tool) Ver. ");
+	cputs("  FD(File & Directory tool) Ver.");
 	cp = strchr(version, ' ');
 	while (*(++cp) == ' ');
 	if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
 	cprintf("%-*.*s", eol - cp, eol - cp, cp);
-	cprintf("%-*.*s", n_column - 34 - (eol - cp),
-		n_column - 34 - (eol - cp), "  (c)1995 T.Shirai  ");
+	cprintf("%-*.*s", n_column - 32 - (eol - cp),
+		n_column - 32 - (eol - cp), " (c)1995,96 T.Shirai  ");
 	putterm(end_standout);
 	timersec = 0;
 	printtime();
@@ -178,7 +335,7 @@ char **cp;
 
 	tmp = *cp;
 	if ((c = **cp) == '"' || c == '\'') {
-		for (*cp = ++tmp; *cp = strchr(*cp, c); *cp++)
+		for (*cp = ++tmp; *cp = strchr(*cp, c); (*cp)++)
 			if (*(*cp - 1) != '\\') {
 				*((*cp)++) = '\0';
 				tmp = strdup2(tmp);
@@ -251,8 +408,9 @@ char *line;
 		free(tmp);
 		tmp = ext;
 		ext = (char *)malloc2(strlen(tmp) + 3);
+		*ext = *tmp;
 		strcpy(ext + 2, tmp);
-		memcpy(ext, "^.*", 3);
+		memcpy(ext + 1, "^.*", 3);
 	}
 	free(tmp);
 
@@ -519,6 +677,68 @@ char *line;
 	return((n) ? 0 : -1);
 }
 
+static int getdosdrive(line)
+char *line;
+{
+	char *cp, *tmp;
+	int i, drive, head, sect, cyl;
+
+	for (i = 0; fdtype[i].name; i++);
+	if (i >= MAXDRIVEENTRY - 1) return(-1);
+	drive = (islower(*line)) ? *line + 'A' - 'a' : *line;
+	cp = skipspace(line + 2);
+	if (!*(tmp = geteostr(&cp))) {
+		free(tmp);
+		return(-1);
+	}
+	cp = skipspace(cp);
+	head = atoi(cp);
+	if (head <= 0 || !(cp = strchr(cp, ','))) {
+		free(tmp);
+		return(-1);
+	}
+	cp = skipspace(cp + 1);
+	sect = atoi(cp);
+	if (sect <= 0 || !(cp = strchr(cp, ','))) {
+		free(tmp);
+		return(-1);
+	}
+	cp = skipspace(cp + 1);
+	cyl = atoi(cp);
+	if (cyl <= 0) {
+		free(tmp);
+		return(-1);
+	}
+	if (*(line + 1) == '!') {
+		for (i = 0; fdtype[i].name; i++)
+		if (drive == fdtype[i].drive
+		&& head == fdtype[i].head
+		&& sect == fdtype[i].sect
+		&& cyl == fdtype[i].cyl
+		&& !strcmp(tmp, fdtype[i].name)) break;
+		free(tmp);
+		if (!fdtype[i].name) return(-1);
+		free(fdtype[i].name);
+		for (; fdtype[i + 1].name; i++) {
+			fdtype[i].drive = fdtype[i + 1].drive;
+			fdtype[i].name = fdtype[i + 1].name;
+			fdtype[i].head = fdtype[i + 1].head;
+			fdtype[i].sect = fdtype[i + 1].sect;
+			fdtype[i].cyl = fdtype[i + 1].cyl;
+		}
+		fdtype[i].name = NULL;
+		return(0);
+	}
+
+	fdtype[i].drive = drive;
+	fdtype[i].name = tmp;
+	fdtype[i].head = head;
+	fdtype[i].sect = sect;
+	fdtype[i].cyl = cyl;
+
+	return(0);
+}
+
 int evalconfigline(line)
 char *line;
 {
@@ -553,6 +773,8 @@ char *line;
 		if (cp = strpbrk(tmp, " \t")) *cp = '\0';
 		loadruncom(tmp);
 	}
+	else if (isalpha(*line) && (*(line + 1) == ':' || *(line + 1) == '!'))
+		 getdosdrive(line);
 	else if (isalpha(*line) || *line == '_') {
 		if (setenv2(line, cp = getenvval(line), 1) < 0) error(line);
 		if (cp) free(cp);
@@ -617,7 +839,7 @@ char *ext;
 {
 	char *cp;
 
-	for (cp = ext; *cp; cp++) {
+	for (cp = ext + 1; *cp; cp++) {
 		if (*cp == '\\') {
 			if (!*(++cp)) break;
 			putch((int)(*(u_char *)cp));
@@ -671,6 +893,20 @@ int printalias()
 	return(n);
 }
 
+int printdrive()
+{
+	int i, n;
+
+	n = 0;
+	for (i = 0; fdtype[i].name; i++) {
+		kanjiprintf("%c:\t\"%s\"\t(%1d,%3d,%3d)\r\n",
+			fdtype[i].drive, fdtype[i].name,
+			fdtype[i].head, fdtype[i].sect, fdtype[i].cyl);
+		if (!(++n % (n_line - 1))) warning(0, HITKY_K);
+	}
+	return(n);
+}
+
 int printhist()
 {
 	int i, n;
@@ -703,8 +939,13 @@ VOID evalenv()
 		minfilename = MINFILENAME;
 	if ((dircountlimit = atoi2(getenv2("FD_DIRCOUNTLIMIT"))) < 0)
 		dircountlimit = DIRCOUNTLIMIT;
+	if ((showsecond = atoi2(getenv2("FD_SECOND"))) < 0) showsecond = SECOND;
+	if ((dosdrive = atoi2(getenv2("FD_DOSDRIVE"))) < 0) dosdrive = DOSDRIVE;
+	if (!(editmode = getenv2("FD_EDITMODE"))) editmode = EDITMODE;
 	if (!(deftmpdir = getenv2("FD_TMPDIR"))) deftmpdir = TMPDIR;
-	language = getenv2("FD_LANGUAGE");
+	deftmpdir = evalpath(strdup2(deftmpdir));
+	inputkcode = getlang(getenv2("FD_INPUTKCODE"), 1);
+	outputkcode = getlang(getenv2("FD_LANGUAGE"), 0);
 }
 
 static int getoption(argc, argv)
@@ -738,20 +979,49 @@ int main(argc, argv)
 int argc;
 char *argv[];
 {
+	char buf[MAXNAMLEN + 1];
 	int i;
 
 	if (progname = strrchr(argv[0], '/')) progname++;
 	else progname = argv[0];
+	sprintf(buf, "%s%d", progname, getpid());
+	tmpfilename = strdup2(buf);
 
 	inittty(0);
 	raw2();
 	noecho2();
 	nonl2();
-	tabs();
+	notabs();
 	getterment();
 	initterm();
 	sigvecset();
 
+	signal(SIGHUP, (sigarg_t (*)())hangup);
+	signal(SIGINT, (sigarg_t (*)())ignore_int);
+	signal(SIGQUIT, (sigarg_t (*)())ignore_quit);
+	signal(SIGILL, (sigarg_t (*)())illerror);
+	signal(SIGTRAP, (sigarg_t (*)())traperror);
+#ifdef	SIGIOT
+	signal(SIGIOT, (sigarg_t (*)())ioerror);
+#endif
+#ifdef	SIGEMT
+	signal(SIGEMT, (sigarg_t (*)())emuerror);
+#endif
+	signal(SIGFPE, (sigarg_t (*)())floaterror);
+	signal(SIGBUS, (sigarg_t (*)())buserror);
+	signal(SIGSEGV, (sigarg_t (*)())segerror);
+#ifdef	SIGSYS
+	signal(SIGSYS, (sigarg_t (*)())syserror);
+#endif
+	signal(SIGPIPE, (sigarg_t (*)())pipeerror);
+	signal(SIGTERM, (sigarg_t (*)())terminate);
+#ifdef	SIGXCPU
+	signal(SIGXCPU, (sigarg_t (*)())xcpuerror);
+#endif
+#ifdef	SIGXFSZ
+	signal(SIGXFSZ, (sigarg_t (*)())xsizerror);
+#endif
+
 	for (maxlaunch = 0; maxlaunch < MAXLAUNCHTABLE; maxlaunch++)
 		if (!launchlist[maxlaunch].ext) break;
 		else {
@@ -771,6 +1041,8 @@ char *argv[];
 				strdup2(archivelist[maxarchive].u_comm);
 		}
 	for (i = 0; i < 10; i++) helpindex[i] = strdup2(helpindex[i]);
+	for (i = 0; fdtype[i].name; i++)
+		fdtype[i].name = strdup2(fdtype[i].name);
 	maxalias = 0;
 
 	loadruncom(DEFRUNCOM);
@@ -799,10 +1071,12 @@ char *argv[];
 		cooked2();
 		echo2();
 		nl2();
+		tabs();
 		inittty(0);
 		raw2();
 		noecho2();
 		nonl2();
+		notabs();
 	}
 
 	sh_history = loadhistory(HISTORYFILE);
@@ -810,6 +1084,7 @@ char *argv[];
 	title();
 
 	main_fd(argv[i]);
+	dosallclose();
 
 	exit2(0);
 }
