@@ -16,5 +16,7 @@ extern char *evalpath();
 extern char *cnvregexp();
 extern reg_t *regexp_init();
 extern int regexp_exec();
-extern void regexp_free();
-extern char *findpath();
+extern int regexp_free();
+extern char *lastpointer();
+extern int completepath();
+extern char *findcommon();
