@@ -7,6 +7,7 @@
 #include "fd.h"
 #include "term.h"
 #include "func.h"
+#include "kctype.h"
 #include "kanji.h"
 
 extern int filepos;
@@ -75,16 +76,25 @@ char *path;
 static char *killmeta(name)
 char *name;
 {
-	char *cp, *buf;
+	u_char *cp, *buf;
 	int i;
+#ifndef	CODEEUC
+	int sjis;
 
-	buf = (char *)malloc2(strlen(name) * 2 + 1);
-	for (cp = name, i = 0; *cp; cp++, i++) {
-		if (strchr(METACHAR, *cp)) buf[i++] = '\\';
+	sjis = (!strcmp((char *)getenv("LANG"), "C")) ? 0 : 1;
+#endif
+
+	buf = (u_char *)malloc2(strlen(name) * 2 + 1);
+	for (cp = (u_char *)name, i = 0; *cp; cp++, i++) {
+#ifndef	CODEEUC
+		if (sjis && iskanji1((int)(*cp))) buf[i++] = *(cp++);
+		else
+#endif
+		if (strchr(METACHAR, (int)(*cp))) buf[i++] = '\\';
 		buf[i] = *cp;
 	}
 	buf[i] = '\0';
-	return(buf);
+	return((char *)buf);
 }
 
 static int setarg(buf, ptr, dir, arg, noext)
@@ -296,6 +306,7 @@ macrostat *stp;
 		for (i = 0; i < max; i++) list[i].flags &= ~F_ISMRK;
 		mark = 0;
 	}
+	if (((stp -> flags) & F_REMAIN) && mark == 0) stp -> flags &= ~F_REMAIN;
 	*(line + j) = '\0';
 	return(evalcomstr(line));
 }
@@ -310,7 +321,7 @@ int max, noconf, argset;
 	int i, status;
 
 	st.flags = (argset) ? F_ARGSET : 0;
-	if (!(tmp = evalcommand(command, arg, list, max, &st))) return(127);
+	if (!(tmp = evalcommand(command, arg, list, max, &st))) return(-1);
 	if (noconf >= 0 && (st.flags & F_NOCONFIRM)) noconf = 1 - noconf;
 	i = (n_column - 4) * WCMDLINE;
 	if (LCMDLINE + WCMDLINE - n_line >= 0) i -= n_column - n_lastcolumn;
@@ -318,9 +329,9 @@ int max, noconf, argset;
 	if (!st.needmark) for (;;) {
 		if (st.addoption >= 0 && noconf >= 0 && !argset
 		&& strlen(tmp) < i) {
-			cp = inputstr2("sh#", st.addoption, tmp, sh_history);
+			cp = inputstr("sh#", 0, st.addoption, tmp, &sh_history);
 			free(tmp);
-			if (!cp || !*cp) return(127);
+			if (!cp || !*cp) return(-1);
 			tmp = evalcomstr(cp);
 			free(cp);
 		}
@@ -334,12 +345,12 @@ int max, noconf, argset;
 	if (mark <= 0) {
 		if (insertarg(buf, tmp, list[filepos].name, st.needmark))
 			status = system3(buf, noconf);
-		else status = 127;
+		else status = -1;
 	}
 	else for (i = 0; i < max; i++) if (ismark(&list[i])) {
 		if (insertarg(buf, tmp, list[i].name, st.needmark))
 			status = system3(buf, noconf);
-		else status = 127;
+		else status = -1;
 	}
 	for (i = 0; i < max; i++) list[i].flags &= ~F_ISMRK;
 	mark = 0;
@@ -372,11 +383,13 @@ int execshell()
 	cooked2();
 	echo2();
 	nl2();
+	tabs();
 	kanjiputs(SHEXT_K);
 	status = system2(sh);
 	raw2();
 	noecho2();
 	nonl2();
+	notabs();
 	sigvecset();
 	putterms(t_keypad);
 	putterms(t_init);
@@ -414,8 +427,9 @@ char **loadhistory(file)
 char *file;
 {
 	FILE *fp;
-	char *cp, line[MAXLINESTR + 1], **hist;
-	int i, j;
+	u_char line[MAXLINESTR + 1];
+	char *cp, **hist;
+	int i, j, len;
 
 	cp = evalpath(strdup2(file));
 	fp = fopen(cp, "r");
@@ -426,10 +440,18 @@ char *file;
 	hist[0] = (char *)histsize;
 
 	i = 1;
-	while (fgets(line, MAXLINESTR, fp)) {
-		if (cp = strchr(line, '\n')) *cp = '\0';
+	while (fgets((char *)line, MAXLINESTR, fp)) {
+		if (cp = strchr((char *)line, '\n')) *cp = '\0';
 		for (j = i; j > 1; j--) hist[j] = hist[j - 1];
-		hist[1] = strdup2(line);
+		for (j = len = 0; line[j]; j++, len++)
+			if (line[j] < ' ' || line[j] == C_DEL) len++;
+		hist[1] = (char *)malloc2(len + 1);
+		for (j = len = 0; line[j]; j++, len++) {
+			if (line[j] < ' ' || line[j] == C_DEL)
+				hist[1][len++] = QUOTE;
+			hist[1][len] = (char)(line[j]);
+		}
+		hist[1][len] = '\0';
 		if (i < histsize) i++;
 	}
 	fclose(fp);
@@ -442,8 +464,8 @@ int savehistory(hist, file)
 char **hist, *file;
 {
 	FILE *fp;
-	char *cp;
-	int i, size;
+	char *cp, line[MAXLINESTR + 1];
+	int i, j, len, size;
 
 	if (!hist || !hist[1]) return(-1);
 	cp = evalpath(strdup2(file));
@@ -453,7 +475,10 @@ char **hist, *file;
 
 	size = (histsize < savehist) ? histsize : savehist;
 	for (i = size; i >= 1; i--) if (hist[i] && *hist[i]) {
-		fputs(hist[i], fp);
+		for (j = 0, len = 0; hist[i][j]; j++)
+			if (hist[i][j] != QUOTE) line[len++] = hist[i][j];
+		line[len] = '\0';
+		fputs(line, fp);
 		fputc('\n', fp);
 	}
 	fclose(fp);
@@ -498,6 +523,7 @@ char *command;
 	cooked2();
 	echo2();
 	nl2();
+	tabs();
 	system2(command);
 	free(command);
 	return(0);
@@ -512,6 +538,7 @@ char *command;
 	raw2();
 	noecho2();
 	nonl2();
+	notabs();
 
 	i = -1;
 	locate(0, n_line - 1);
@@ -528,6 +555,7 @@ char *command;
 	else if (!strcmp(command, "printlaunch")) i = printlaunch();
 	else if (!strcmp(command, "printarch")) i = printarch();
 	else if (!strcmp(command, "alias")) i = printalias();
+	else if (!strcmp(command, "printdrive")) i = printdrive();
 	else if (!strcmp(command, "history")) i = printhist();
 	else if (!strncmp(command, "cd", cp - command)) {
 		while (*cp == ' ' || *cp == '\t') cp++;
@@ -583,24 +611,24 @@ char *command;
 	return(cp);
 }
 
-int completealias(com, match)
-char *com, *match;
+int completealias(com, matchno, matchp)
+char *com;
+int matchno;
+char **matchp;
 {
-	int i, j, len, matchno;
+	int i, len, ptr, size;
 
 	if (strchr(com, '/')) return(0);
 
-	matchno = 0;
+	size = lastpointer(*matchp, matchno) - *matchp;
 	len = strlen(com);
 	for (i = 0; i < maxalias; i++) {
 		if (strncmp(com, aliaslist[i].alias, len)) continue;
-		if (!matchno++) strcpy(match, aliaslist[i].alias);
-		else {
-			for (j = 0; match[j]; j++)
-				if (match[j] != aliaslist[i].alias[j]) break;
-			match[j] = '\0';
-		}
+		ptr = size;
+		size += strlen(aliaslist[i].alias) + 1;
+		*matchp = (char *)realloc2(*matchp, size);
+		strcpy(*matchp + ptr, aliaslist[i].alias);
+		matchno++;
 	}
-	if (!matchno) *match = '\0';
 	return(matchno);
 }
