@@ -4,6 +4,7 @@
  *	Terminal Module
  */
 
+#include "machine.h"
 #include <stdio.h>
 #include <string.h>
 #include <varargs.h>
@@ -11,8 +12,6 @@
 #include <sys/types.h>
 #include <sys/file.h>
 #include <sys/time.h>
-#include "machine.h"
-#include "term.h"
 
 #ifndef	NOSTDLIBH
 #include <stdlib.h>
@@ -28,6 +27,8 @@
 #include <sys/select.h>
 #endif
 
+#include "term.h"
+
 extern int tgetent();
 extern int tgetnum();
 extern int tgetflag();
@@ -48,7 +49,25 @@ extern char *tgoto();
 #define	GETSIZE		"\0337\033[r\033[999;999H\033[6n"
 #define	SIZEFMT		"\033[%d;%dR"
 
-#ifndef FD_SET
+#ifndef	PENDIN
+#define	PENDIN		0
+#endif
+#ifndef	ECHOCTL
+#define	ECHOCTL		0
+#endif
+#ifndef	ECHOKE
+#define	ECHOKE		0
+#endif
+
+#ifndef	FREAD
+# ifdef	_FREAD
+# define	FREAD	_FREAD
+# else
+# define	FREAD	(O_RDONLY + 1)
+# endif
+#endif
+
+#ifndef	FD_SET
 typedef struct fd_set {
 	int fds_bits[1];
 } fd_set;
@@ -109,13 +128,14 @@ char *c_right;
 char *c_left;
 
 static char *keyseq[K_MAX - K_MIN + 1];
-static unsigned char keyseqlen[K_MAX - K_MIN + 1];
+static u_char keyseqlen[K_MAX - K_MIN + 1];
 static short keycode[K_MAX - K_MIN + 1];
 static int ttyio;
+static FILE *ttyout;
 static char *termname;
 
 #ifndef	TIOCSTI
-static unsigned char ungetbuf[10];
+static u_char ungetbuf[10];
 static int ungetnum = 0;
 #endif
 
@@ -128,17 +148,22 @@ static int termflags;
 int inittty(reset)
 int reset;
 {
-	unsigned long request;
 #ifdef	USETERMIO
 	static struct termio dupttyio;
-
-	request = (reset) ? TCSETAW : TCGETA;
 #else
 	static struct sgttyb dupttyio;
+	static int dupttyflag;
+#endif
+	u_long request;
 
+	if (reset && !(termflags & F_INITTTY)) return;
+#ifdef	USETERMIO
+	request = (reset) ? TCSETAW : TCGETA;
+#else
+	request = (reset) ? TIOCLSET : TIOCLGET;
+	if (ioctl(ttyio, request, &dupttyflag) < 0) err2(NULL);
 	request = (reset) ? TIOCSETP : TIOCGETP;
 #endif
-	if (reset && !(termflags & F_INITTTY)) return;
 	if (ioctl(ttyio, request, &dupttyio) < 0) err2(NULL);
 	if (!reset) {
 #ifdef	USETERMIO
@@ -150,153 +175,127 @@ int reset;
 	}
 }
 
-static int ttymode(d, mode, set)
-int d;
-short mode;
-int set;
-{
 #ifdef	USETERMIO
+static int ttymode(d, set, reset, iset, ireset, oset, oreset, min, time)
+int d;
+u_short set, reset, iset, ireset, oset, oreset;
+int min, time;
+{
 	struct termio tty;
 
 	if (ioctl(d, TCGETA, &tty) < 0) err2(NULL);
-	if (set) tty.c_lflag |= mode;
-	else tty.c_lflag &= ~mode;
+	if (set) tty.c_lflag |= set;
+	if (reset) tty.c_lflag &= reset;
+	if (iset) tty.c_iflag |= iset;
+	if (ireset) tty.c_iflag &= ireset;
+	if (oset) tty.c_oflag |= oset;
+	if (oreset) tty.c_oflag &= oreset;
+	if (min) {
+		tty.c_cc[VMIN] = min;
+		tty.c_cc[VTIME] = time;
+	}
 	if (ioctl(d, TCSETAW, &tty) < 0) err2(NULL);
 #else
+static int ttymode(d, set, reset, lset, lreset)
+int d;
+u_short set, reset, lset, lreset;
+{
 	struct sgttyb tty;
+	int lflag;
 
 	if (ioctl(d, TIOCGETP, &tty) < 0) err2(NULL);
-	if (set) tty.sg_flags |= mode;
-	else tty.sg_flags &= ~mode;
+	if (ioctl(d, TIOCLGET, &lflag) < 0) err2(NULL);
+	if (set) tty.sg_flags |= set;
+	if (reset) tty.sg_flags &= reset;
+	if (set) lflag |= lset;
+	if (lreset) lflag &= lreset;
 	if (ioctl(d, TIOCSETP, &tty) < 0) err2(NULL);
+	if (ioctl(d, TIOCLSET, &lflag) < 0) err2(NULL);
 #endif
 }
 
 int cooked2()
 {
 #ifdef	USETERMIO
-	struct termio tty;
-
-	if (ioctl(ttyio, TCGETA, &tty) < 0) err2(NULL);
-	tty.c_iflag &= ~IGNBRK;
-	tty.c_iflag |= IXON;
-	tty.c_oflag |= OPOST;
-	tty.c_oflag &= ~(OCRNL|ONOCR|ONLRET);
-	tty.c_lflag |= ISIG|ICANON;
-	tty.c_cc[VMIN] = 4;
-	tty.c_cc[VTIME] = 255;
-	if (ioctl(ttyio, TCSETAW, &tty) < 0) err2(NULL);
+	ttymode(ttyio, ISIG | ICANON, ~PENDIN,
+		BRKINT | IGNPAR | IXON | IXANY | IXOFF, ~IGNBRK,
+		OPOST, 0, '\004', 255);
 #else
-	ttymode(ttyio, CBREAK | RAW, 0);
+	ttymode(ttyio, 0, ~(CBREAK | RAW), 0, ~(LLITOUT | LPENDIN));
 #endif
 }
 
 int cbreak2()
 {
 #ifdef	USETERMIO
-	struct termio tty;
-
-	if (ioctl(ttyio, TCGETA, &tty) < 0) err2(NULL);
-	tty.c_iflag &= ~IGNBRK;
-	tty.c_iflag &= ~IXON;
-	tty.c_oflag |= OPOST;
-	tty.c_oflag &= ~(OCRNL|ONOCR|ONLRET);
-	tty.c_lflag |= ISIG|ICANON;
-	tty.c_cc[VMIN] = 1;
-	tty.c_cc[VTIME] = 0;
-	if (ioctl(ttyio, TCSETAW, &tty) < 0) err2(NULL);
+	ttymode(ttyio, ISIG | ICANON, 0, BRKINT, ~(IXON | IGNBRK),
+		0, ~OPOST, 1, 0);
 #else
-	ttymode(ttyio, CBREAK, 1);
+	ttymode(ttyio, CBREAK, 0, LLITOUT, 0);
 #endif
 }
 
 int raw2()
 {
 #ifdef	USETERMIO
-	struct termio tty;
-
-	if (ioctl(ttyio, TCGETA, &tty) < 0) err2(NULL);
-	tty.c_iflag |= IGNBRK;
-	tty.c_iflag &= ~IXON;
-	tty.c_oflag &= ~OPOST;
-	tty.c_lflag &= ~(ISIG|ICANON);
-	tty.c_cc[VMIN] = 1;
-	tty.c_cc[VTIME] = 0;
-	if (ioctl(ttyio, TCSETAW, &tty) < 0) err2(NULL);
+	ttymode(ttyio, 0, ~(ISIG | ICANON), IGNBRK, ~IXON, 0, ~OPOST, 1, 0);
 #else
-	ttymode(ttyio, RAW, 1);
+	ttymode(ttyio, RAW, 0, LLITOUT, 0);
 #endif
 }
 
 int echo2()
 {
 #ifdef	USETERMIO
-	ttymode(ttyio, ECHO|ECHOE|ECHOK|ECHONL, 1);
+	ttymode(ttyio, ECHO | ECHOE | ECHOK | ECHOCTL | ECHOKE, ~ECHONL,
+		0, 0, 0, 0, 0, 0);
 #else
-	ttymode(ttyio, ECHO, 1);
+	ttymode(ttyio, ECHO, 0, LCRTBS | LCRTERA | LCRTKIL | LCTLECH, 0);
 #endif
 }
 
 int noecho2()
 {
 #ifdef	USETERMIO
-	ttymode(ttyio, ECHO|ECHOE|ECHOK|ECHONL, 0);
+	ttymode(ttyio, 0, ~(ECHO | ECHOE | ECHOK | ECHONL), 0, 0, 0, 0, 0, 0);
 #else
-	ttymode(ttyio, ECHO, 0);
+	ttymode(ttyio, 0, ~ECHO, 0, ~(LCRTBS | LCRTERA));
 #endif
 }
 
 int nl2()
 {
 #ifdef	USETERMIO
-	struct termio tty;
-
-	if (ioctl(ttyio, TCGETA, &tty) < 0) err2(NULL);
-	tty.c_iflag |= ICRNL;
-	tty.c_oflag |= ONLCR;
-	if (ioctl(ttyio, TCSETAW, &tty) < 0) err2(NULL);
+	ttymode(ttyio, 0, 0, ICRNL, 0, ONLCR, ~(OCRNL | ONOCR | ONLRET), 0, 0);
 #else
-	ttymode(ttyio, CRMOD, 1);
+	ttymode(ttyio, CRMOD, 0, 0, 0);
 #endif
 }
 
 int nonl2()
 {
 #ifdef	USETERMIO
-	struct termio tty;
-
-	if (ioctl(ttyio, TCGETA, &tty) < 0) err2(NULL);
-	tty.c_iflag &= ~ICRNL;
-	tty.c_oflag &= ~ONLCR;
-	if (ioctl(ttyio, TCSETAW, &tty) < 0) err2(NULL);
+	ttymode(ttyio, 0, 0, 0, ~ICRNL, 0, ~ONLCR, 0, 0);
 #else
-	ttymode(ttyio, CRMOD, 0);
+	ttymode(ttyio, 0, ~CRMOD, 0, 0);
 #endif
 }
 
 int tabs()
 {
 #ifdef	USETERMIO
-	struct termio tty;
-
-	if (ioctl(ttyio, TCGETA, &tty) < 0) err2(NULL);
-	tty.c_oflag &= ~TAB3;
-	if (ioctl(ttyio, TCSETAW, &tty) < 0) err2(NULL);
+	ttymode(ttyio, 0, 0, 0, 0, 0, ~TAB3, 0, 0);
 #else
-	ttymode(ttyio, XTABS, 0);
+	ttymode(ttyio, 0, ~XTABS, 0, 0);
 #endif
 }
 
 int notabs()
 {
 #ifdef	USETERMIO
-	struct termio tty;
-
-	if (ioctl(ttyio, TCGETA, &tty) < 0) err2(NULL);
-	tty.c_oflag |= TAB3;
-	if (ioctl(ttyio, TCSETAW, &tty) < 0) err2(NULL);
+	ttymode(ttyio, 0, 0, 0, 0, TAB3, 0, 0, 0);
 #else
-	ttymode(ttyio, XTABS, 1);
+	ttymode(ttyio, XTABS, 0, 0, 0);
 #endif
 }
 
@@ -323,12 +322,13 @@ int no;
 static int err2(mes)
 char *mes;
 {
+	if (termflags & F_TERMENT) putterm(t_normal);
 	endterm();
 	inittty(1);
-	cprintf("\007\n");
-	if (mes) cprintf("%s\n", mes);
+	fprintf(stderr, "\007\n");
+	if (mes) fprintf(stderr, "%s\n", mes);
 	else perror(TTYNAME);
-	exit2(1);
+	exit(127);
 }
 
 static int defaultterm()
@@ -336,9 +336,11 @@ static int defaultterm()
 	int i;
 
 	BC = "\010";
-	UP = "\033[A";
+	UP = "2\033[A";
 	n_column = n_lastcolumn = 80;
 	n_line = 24;
+	t_init = "";
+	t_end = "";
 	t_keypad = "\033[?1h\033=";
 	t_nokeypad = "\033[?1l\033>";
 	t_normalcursor = "\033[?25h";
@@ -348,17 +350,17 @@ static int defaultterm()
 	t_resetcursor = "\0338";
 	t_bell = "\007";
 	t_vbell = "\007";
-	t_clear = "\033[2J";
-	t_normal = "\033[m";
-	t_bold = "\033[1m";
-	t_reverse = "\033[7m";
-	t_dim = "\033[2m";
-	t_blink = "\033[5m";
-	t_standout = "\033[7m";
-	t_underline = "\033[4m";
-	end_standout = "\033[m";
-	end_underline = "\033[m";
-	l_clear = "\033[K";
+	t_clear = "50\033[;H\033[2J";
+	t_normal = "2\033[m";
+	t_bold = "2\033[1m";
+	t_reverse = "2\033[7m";
+	t_dim = "2\033[2m";
+	t_blink = "2\033[5m";
+	t_standout = "2\033[7m";
+	t_underline = "2\033[4m";
+	end_standout = "2\033[m";
+	end_underline = "2\033[m";
+	l_clear = "3\033[K";
 	l_insert = "\033[L";
 	l_delete = "\033[M";
 	c_insert = "";
@@ -366,12 +368,12 @@ static int defaultterm()
 	c_store = "\033[s";
 	c_restore = "\033[u";
 	c_home = "\033[H";
-	c_locate = "\033[%d;%dH";
+	c_locate = "5\033[%i%d;%dH";
 	c_return = "\r";
 	c_newline = "\n";
-	c_up = "\033[A";
+	c_up = "2\033[A";
 	c_down = "\012";
-	c_right = "\033[C";
+	c_right = "2\033[C";
 	c_left = "\010";
 
 	for (i = 0; i <= K_MAX - K_MIN; i++) keyseq[i] = NULL;
@@ -379,6 +381,7 @@ static int defaultterm()
 	keyseq[K_DOWN - K_MIN] = "\033OB";
 	keyseq[K_RIGHT - K_MIN] = "\033OC";
 	keyseq[K_LEFT - K_MIN] = "\033OD";
+	keyseq[K_HOME - K_MIN] = "\033[1~";
 	keyseq[K_BS - K_MIN] = "\010";
 	keyseq[K_F(1) - K_MIN] = "\033[11~";
 	keyseq[K_F(2) - K_MIN] = "\033[12~";
@@ -429,10 +432,15 @@ static int defaultterm()
 	keyseq[K_F('9') - K_MIN] = "\033Oy";
 	keyseq[K_F('=') - K_MIN] = "\033OX";
 	keyseq[K_F('?') - K_MIN] = "\033OM";
+	keyseq[K_DL - K_MIN] = "";
+	keyseq[K_IL - K_MIN] = "";
 	keyseq[K_DC - K_MIN] = "\177";
 	keyseq[K_IC - K_MIN] = "\033[2~";
+	keyseq[K_CLR - K_MIN] = "";
 	keyseq[K_PPAGE - K_MIN] = "\033[5~";
 	keyseq[K_NPAGE - K_MIN] = "\033[6~";
+	keyseq[K_ENTER - K_MIN] = "\033[9~";
+	keyseq[K_END - K_MIN] = "\033[4~";
 }
 
 static int tgetstr2(term, str)
@@ -515,7 +523,8 @@ int getterment()
 	if (!(termname = (char *)getenv("TERM"))) termname = "unknown";
 	if (tgetent(buf, termname) <= 0) err2("No TERMCAP prepared");
 
-	if ((ttyio = open(TTYNAME, O_RDWR)) < 0) ttyio = STDERR;
+	if ((ttyio = open(TTYNAME, O_RDWR, 0600)) < 0) ttyio = STDERR;
+	if (!(ttyout = fopen(TTYNAME, "w+"))) ttyout = stdout;
 
 	defaultterm();
 	cp = "";
@@ -569,6 +578,7 @@ int getterment()
 	tgetstr2(&keyseq[K_DOWN - K_MIN], "kd");
 	tgetstr2(&keyseq[K_RIGHT - K_MIN], "kr");
 	tgetstr2(&keyseq[K_LEFT - K_MIN], "kl");
+	tgetstr2(&keyseq[K_HOME - K_MIN], "kh");
 	tgetstr2(&keyseq[K_BS - K_MIN], "kb");
 	tgetstr2(&keyseq[K_F(1) - K_MIN], "l1");
 	tgetstr2(&keyseq[K_F(2) - K_MIN], "l2");
@@ -590,9 +600,17 @@ int getterment()
 	tgetstr2(&keyseq[K_F(28) - K_MIN], "k8");
 	tgetstr2(&keyseq[K_F(29) - K_MIN], "k9");
 	tgetstr2(&keyseq[K_F(30) - K_MIN], "k0");
+	tgetstr2(&keyseq[K_DL - K_MIN], "kL");
+	tgetstr2(&keyseq[K_IL - K_MIN], "kA");
 	tgetstr2(&keyseq[K_DC - K_MIN], "kD");
+	tgetstr2(&keyseq[K_IC - K_MIN], "kI");
+	tgetstr2(&keyseq[K_CLR - K_MIN], "kC");
+	tgetstr2(&keyseq[K_EOL - K_MIN], "kE");
 	tgetstr2(&keyseq[K_PPAGE - K_MIN], "kP");
 	tgetstr2(&keyseq[K_NPAGE - K_MIN], "kN");
+	tgetstr2(&keyseq[K_ENTER - K_MIN], "@8");
+	tgetstr2(&keyseq[K_BEG - K_MIN], "@1");
+	tgetstr2(&keyseq[K_END - K_MIN], "@7");
 	sortkeyseq();
 
 	termflags |= F_TERMENT;
@@ -619,13 +637,13 @@ int endterm()
 int putch(c)
 int c;
 {
-	return(fputc(c, stdout));
+	return(fputc(c, ttyout));
 }
 
 int cputs(str)
 char *str;
 {
-	return(fputs(str, stdout));
+	return(fputs(str, ttyout));
 }
 
 #ifndef	NOVSPRINTF
@@ -651,12 +669,12 @@ char *fmt;
 	len = sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6);
 #endif
 
-	fputs(buf, stdout);
+	fputs(buf, ttyout);
 	return(len);
 }
 
 int kbhit2(usec)
-unsigned long usec;
+u_long usec;
 {
 #ifdef	NOSELECT
 	return((usec) ? 1 : 0);
@@ -669,15 +687,15 @@ unsigned long usec;
 	FD_ZERO(&readfds);
 	FD_SET(STDIN, &readfds);
 
-	return (select(1, &readfds, NULL, NULL, &timeout));
+	return(select(1, &readfds, NULL, NULL, &timeout));
 #endif
 }
 
 int getch2()
 {
-	unsigned char ch;
+	u_char ch;
 
-	read(ttyio, &ch, sizeof(unsigned char));
+	read(ttyio, &ch, sizeof(u_char));
 	return((int)ch);
 }
 
@@ -716,7 +734,7 @@ int sig;
 }
 
 int ungetch2(c)
-unsigned char c;
+u_char c;
 {
 #ifdef	TIOCSTI
 	ioctl(ttyio, TIOCSTI, &c);
@@ -734,7 +752,7 @@ int x, y;
 
 int tflush()
 {
-	fflush(stdout);
+	fflush(ttyout);
 }
 
 static int realscanf(s, format, yp, xp)
