@@ -9,10 +9,10 @@ typedef u_short	gid_t;
 
 #ifdef	NOVOID
 #define	VOID
-#define VOID_P	char *
+#define	VOID_P	char *
 #else
-#define VOID	void
-#define VOID_P	void *
+#define	VOID	void
+#define	VOID_P	void *
 #endif
 
 #ifdef	USEDIRECT
@@ -143,3 +143,6 @@ typedef struct _aliastable {
 #define	isdisplnk(n)		((n) & F_SYMLINK)
 #define	isdisptyp(n)		((n) & F_FILETYPE)
 #define	ishidedot(n)		((n) & F_DOTFILE)
+
+#define	QUOTE	('^' - '@')
+#define	C_DEL	'\177'
